# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Populatiron RL (PopRL) algorithm for repeated rock-paper-scissors.

For details, see Lanctot et al. 2023 Population-based Evaluation in Repeated
Rock-Paper-Scissors as a Benchmark for Multiagent Reinforcement Learning
https://openreview.net/forum?id=gQnJ7ODIAx
"""

import copy
import sys
import time

from absl import app
from absl import flags
import jax
import numpy as np

from open_spiel.python import rl_agent
from open_spiel.python import rl_tools
from open_spiel.python.algorithms import random_agent
from open_spiel.python.algorithms import tabular_qlearner
from open_spiel.python.examples.rrps_poprl import impala
from open_spiel.python.examples.rrps_poprl import rl_environment
from open_spiel.python.jax import boltzmann_dqn
from open_spiel.python.jax import dqn
from open_spiel.python.jax import policy_gradient
import pyspiel

FLAGS = flags.FLAGS

# Training parameters
flags.DEFINE_string(
    "checkpoint_dir",
    "/tmp/dqn_test",
    "Directory to save/load the agent models.",
)
flags.DEFINE_integer(
    "save_every",
    int(1e4),
    "Episode frequency at which the DQN agent models are saved.",
)
flags.DEFINE_integer(
    "num_train_episodes", int(1e6), "Number of training episodes."
)
flags.DEFINE_integer(
    "eval_every",
    100,
    "Episode frequency at which the DQN agents are evaluated.",
)
flags.DEFINE_integer("eval_episodes", 1, "How many episodes to run per eval.")

# DQN model hyper-parameters
flags.DEFINE_list(
    "hidden_layers_sizes",
    [256, 128],
    "Number of hidden units in the Q-Network MLP.",
)
flags.DEFINE_integer(
    "replay_buffer_capacity", int(1e5), "Size of the replay buffer."
)
flags.DEFINE_integer(
    "batch_size", 16, "Number of transitions to sample at each learning step."
)
flags.DEFINE_float("learning_rate", 0.01, "Learning rate.")
flags.DEFINE_float("eta", 0.2, "BDQN eta param")

# policy_gradient params
flags.DEFINE_float("critic_learning_rate", 0.001, "Critic Learning rate.")
flags.DEFINE_float("pi_learning_rate", 0.0001, "Pi learning rate.")
flags.DEFINE_float("entropy_cost", 0.001, "Entropy cost.")
flags.DEFINE_float("lambda_", 1.0, "PG lambda.")
flags.DEFINE_integer("num_critic_before_pi", 8, "Entropy cost.")

# impala params
flags.DEFINE_integer("unroll_length", 20, "Unroll length.")
flags.DEFINE_float(
    "prediction_weight", 0, "Weight to put on the prediction losses."
)

# Main algorithm parameters
flags.DEFINE_integer("seed", 0, "Seed to use for everything")
flags.DEFINE_integer("window_size", 50, "Size of window for rolling average")
flags.DEFINE_integer("num_players", 2, "Numebr of players")
flags.DEFINE_string("game", "leduc_poker", "Game string")
flags.DEFINE_string("exploitee", "random", "Exploitee (random | first)")
flags.DEFINE_string("learner", "impala", "Learner (qlearning | dqn)")

flags.DEFINE_integer("cp_freq", 10000, "Checkpoint save frequency.")
flags.DEFINE_string("cp_dir", None, "Checkpoint directory")

# Testing against specific members
flags.DEFINE_integer("pop_only", -1, "Create a population of only this bot.")

# Generalization. How many agents to leave ou of the training population and
# use for testing?
flags.DEFINE_integer("leave_out_set_size", 0, "Cross-validation test size.")

# Environment recall
flags.DEFINE_integer("env_recall", 1, "How many timesteps back define obs?")

flags.DEFINE_string("pred_logs_dir", None, "Directory to save prediction logs.")

# Interactive mode
flags.DEFINE_string("interactive_mode", None, 'Bot id or "human".')

flags.DEFINE_float("rm_epsilon", 0.1, "Exploration for regret-matching.")

# Population RL
flags.DEFINE_float("prob_selfplay", 0.2, "Probability that we meet ourself")
flags.DEFINE_string("eval_checkpoint", None, "Evaluate a checkpoint")

# Set this to something specific for testing. List of IDs
# Set back to None to use full population.
FIXED_POPULATION = None


class State:

  def __init__(self):
    self.np_rng_state = None
    self.learning_agents = None
    self.ep = None
    self.rolling_averager = None
    self.expl_rolling_averagers = None


## This does not work, unfortunately. Not sure why. Simple pickle does not work
## because one of the haiku transforms is not serializable. There seems to be
## some nontrivial logic to use checkpointing when working with Haiku.
## Below is my attempt at applying a fix I found based on this thread:
## https://github.com/google-deepmind/dm-haiku/issues/18 but it didn't work.
class Checkpoint(object):
  """A class for saving the state of the agent (and model)."""

  def __init__(self, checkpoint_dir):
    self.checkpoint_dir = checkpoint_dir
    self.state = State()

  def restore_or_save(self):
    assert False, "Not implemented yet."
    # filename = os.path.join(self.checkpoint_dir, "tree.pkl")
    # if os.path.exists(filename):
    #   self.state = self.restore()
    # else:
    #   # pickle.dump(self.state, filename)  # Pickles to any file (even /cns).
    #   self.save()

  def restore(self):
    assert False, "Not implemented yet."
    # print("Restoring checkpoint")
    # with open(os.path.join(self.checkpoint_dir, "tree.pkl"), "rb") as f:
    #   tree_struct = pickle.load(f)
    # leaves, treedef = jax.tree_util.tree_flatten(tree_struct)
    # with open(os.path.join(self.checkpoint_dir, "arrays.npy"), "rb") as f:
    #   flat_state = [np.load(f, allow_pickle=False) for _ in leaves]
    # return jax.tree_util.tree_unflatten(treedef, flat_state)

  def save(self):
    assert False, "Not implemented yet."
    # print("Saving checkpoint")
    # # filename = os.path.join(self.checkpoint_dir, "checkpoint.pkl")
    # # pickle.dump(self.state, filename)  # Pickles to any file (even /cns).
    # with open(os.path.join(self.checkpoint_dir, "arrays.npy"), "wb") as f:
    #   for x in jax.tree_util.tree_leaves(self.state):
    #     np.save(f, x, allow_pickle=False)
    # tree_struct = jax.tree_util.tree_map(lambda t: 0, self.state)
    # with open(os.path.join(self.checkpoint_dir, "tree.pkl"), "wb") as f:
    #   pickle.dump(tree_struct, f)


class PredictionLogger(object):
  """A prediction logger."""

  def __init__(self, log_dir):
    self._log_dir = log_dir
    self._enabled = self._log_dir is not None
    self._logs = {}
    self._cur_log = ""
    self._freq = 1000
    self._last_log = 0

  def new_log(self, training_episodes):
    if not self._enabled:
      return
    if training_episodes - self._last_log >= self._freq:
      self._cur_log = ""
      self._cur_step = 0

  def log(self, training_episodes, pop_idx, predictions):
    if not self._enabled:
      return
    if training_episodes - self._last_log >= self._freq:
      line = f"{training_episodes} {pop_idx} {self._cur_step}"
      for i in range(len(predictions)):
        line = line + f" {predictions[i]}"
      self._cur_log += line + "\n"
      self._cur_step += 1

  def end_log(self, training_episodes, pop_idx):
    if not self._enabled:
      return
    if training_episodes - self._last_log >= self._freq:
      key = f"{training_episodes}.{pop_idx}"
      self._logs[key] = self._cur_log

  def update_training_episodes(self, training_episodes):
    if not self._enabled:
      return
    if training_episodes - self._last_log >= self._freq:
      self._last_log = training_episodes


def last_predictions(agent):
  if hasattr(agent, "last_predictions"):
    return agent.last_predictions()
  else:
    return np.zeros(pyspiel.ROSHAMBO_NUM_BOTS)


def eval_agent(
    env,
    num_players,
    num_actions,
    bot_names,
    learning_agent,
    prediction_logger,
    num_training_episodes,
):
  """Evaluate the agent."""
  sum_episode_rewards = np.zeros(num_players)
  pop_expl = np.zeros(pyspiel.ROSHAMBO_NUM_BOTS)
  for pop_idx in range(len(bot_names)):
    bot_id = pop_idx
    bot_name = bot_names[bot_id]
    bot = pyspiel.make_roshambo_bot(0, bot_name)
    pop_agent = BotAgent(num_actions, bot, name=bot_name)

    if hasattr(learning_agent, "restart"):
      learning_agent.restart()

    agents = [pop_agent, learning_agent]
    env.set_prediction_label(pop_idx)

    time_step = env.reset()
    episode_rewards = np.zeros(num_players)
    turn_num = 0
    prediction_logger.new_log(num_training_episodes)

    while not time_step.last():
      turn_num += 1
      player_id = time_step.observations["current_player"]
      if env.is_turn_based:
        agent_output = agents[player_id].step(time_step, is_evaluation=True)
        action_list = [agent_output.action]
      else:
        agents_output = [
            agent.step(time_step, is_evaluation=True) for agent in agents
        ]
        action_list = [agent_output.action for agent_output in agents_output]
      prediction_logger.log(
          num_training_episodes, pop_idx, last_predictions(learning_agent)
      )
      time_step = env.step(action_list)
      episode_rewards += time_step.rewards
    pop_expl[pop_idx] = episode_rewards[0]
    sum_episode_rewards += episode_rewards
    prediction_logger.end_log(num_training_episodes, pop_idx)
  prediction_logger.update_training_episodes(num_training_episodes)
  return sum_episode_rewards / len(bot_names), pop_expl


class HumanAgent(rl_agent.AbstractAgent):
  """Agent class that wraps a bot.

  Note, the environment must include the OpenSpiel state.
  """

  def __init__(self, num_actions, name="human_agent"):
    assert num_actions > 0
    self._num_actions = num_actions

  def step(self, time_step, is_evaluation=False):
    action = 5
    while action > 2:
      value_str = input("Choose an action: ")
      if value_str == "R":
        action = 0
      if value_str == "P":
        action = 1
      if value_str == "S":
        action = 2
      if value_str == "q":
        action = -1
    probs = np.zeros(self._num_actions)
    if action >= 0:
      probs[action] = 1.0
    return rl_agent.StepOutput(action=action, probs=probs)


def pretty_top10_preds_str(predictions, indices, max_weight=1.01):
  """Pretty string representation of the top 10 predictions."""

  top_10_preds = ""
  sum_weight = 0
  for i in range(10):
    pred_idx = indices[42 - i]
    weight = predictions[pred_idx]
    bar_width = int(weight / 0.01)
    bar_str = "#" * bar_width
    top_10_preds += f"  {pred_idx:2d}: {weight:.5f} {bar_str}\n"
    sum_weight += weight
    if sum_weight > max_weight:
      break
  return top_10_preds


def interactive_episode(
    env, num_players, num_actions, bot_names, learning_agent
):
  """Interactive Episode."""
  print("Starting interactive episode!")
  actions_str = ["R", "P", "S"]
  actions_seq = ["", ""]

  if FLAGS.interactive_mode == "human":
    pop_agent = HumanAgent(num_actions)
    pop_idx = -1
  else:
    test_pop_ids = [int(FLAGS.interactive_mode)]
    pop_agent, pop_idx = sample_bot_agent(bot_names, test_pop_ids, num_actions)
    print(f"Sampled bot {pop_idx} ({bot_names[pop_idx]})")

  agents = [pop_agent, learning_agent]

  time_step = env.reset()
  episode_rewards = np.zeros(num_players)
  turn_num = 0

  while not time_step.last():
    player_id = time_step.observations["current_player"]
    if env.is_turn_based:
      agent_output = agents[player_id].step(time_step, is_evaluation=True)
      action_list = [agent_output.action]
    else:
      agents_output = [
          agent.step(time_step, is_evaluation=True) for agent in agents
      ]
      action_list = [agent_output.action for agent_output in agents_output]
    if action_list[0] == -1:
      # Restart episode.
      print("Restarting episode.")
      interactive_episode(
          env, num_players, num_actions, bot_names, learning_agent
      )
      return
    action_list_str = [actions_str[int(x)] for x in action_list]
    actions_seq[0] += action_list_str[0]
    actions_seq[1] += action_list_str[1]
    predictions = last_predictions(learning_agent)
    indices = np.argsort(predictions)
    top_10_preds = pretty_top10_preds_str(predictions, indices, max_weight=0.75)
    time_step = env.step(action_list)
    episode_rewards += time_step.rewards
    print(
        f"Turn {turn_num}, Prev actions: {action_list_str}, "
        + f"Rewards: {time_step.rewards}, Returns: {episode_rewards} \n"
        + f"Action Seq [0]: {actions_seq[0]} \n"
        + f"Action Seq [1]: {actions_seq[1]}"
    )
    print(f"Top 10 predictions: \n{top_10_preds}")
    if FLAGS.interactive_mode != "human":
      # Prompt to continue.
      input("Press any key:")
    turn_num += 1


class ConstantActionAgent(rl_agent.AbstractAgent):
  """An example agent class."""

  def __init__(
      self, player_id, num_actions, action_idx, name="constant_action_agent"
  ):
    assert num_actions > 0
    self._player_id = player_id
    self._num_actions = num_actions
    self._action_idx = action_idx

  def step(self, time_step, is_evaluation=False):
    # If it is the end of the episode, don't select an action.
    if time_step.last():
      return

    cur_legal_actions = time_step.observations["legal_actions"][self._player_id]
    action = cur_legal_actions[self._action_idx]
    probs = np.zeros(self._num_actions)
    probs[action] = 1.0
    return rl_agent.StepOutput(action=action, probs=probs)


class RegretMatchingAgent(rl_agent.AbstractAgent):
  """TODO(author5): finish this agent."""

  def __init__(
      self,
      player_id,
      num_actions,
      epsilon,
      constant_observation=None,
      name="regret_matching_agent",
  ):
    assert num_actions > 0
    self._player_id = player_id
    self._num_actions = num_actions
    self._regrets = {}
    self._prev_info_state = None
    self._prev_action = None
    self._prev_sample_policy = None
    self._prev_rm_policy = None
    self._epsilon = epsilon
    self._prev_legal_actions = None
    self._constant_observation = constant_observation

  def _get_info_state_key(self, info_state):
    return (
        self._constant_observation
        if self._constant_observation is not None
        else info_state
    )

  def _get_rm_policy(self, uniform_policy, info_state, legal_actions):
    info_state_key = self._get_info_state_key(info_state)
    regrets = self._regrets.get(info_state_key, None)
    if regrets is None:
      regrets = np.zeros(self._num_actions, dtype=np.float64)
      regrets[legal_actions] = 0.000001
      self._regrets[info_state_key] = regrets
    rm_policy = regrets.copy()
    rm_policy[rm_policy < 0] = 0.0
    denom = rm_policy.sum()
    if denom <= 0:
      rm_policy = uniform_policy
    else:
      rm_policy /= denom
    return rm_policy

  def _get_action_probs(self, info_state, legal_actions, epsilon):
    uniform_policy = np.zeros(self._num_actions, dtype=np.float64)
    uniform_policy[legal_actions] = 1.0 / len(legal_actions)
    rm_policy = self._get_rm_policy(uniform_policy, info_state, legal_actions)
    sample_policy = epsilon * uniform_policy + (1 - epsilon) * rm_policy
    # print(sample_policy)
    action = np.random.choice(np.arange(self._num_actions), p=sample_policy)
    return action, sample_policy, rm_policy

  def step(self, time_step, is_evaluation=False):
    legal_actions = time_step.observations["legal_actions"][self._player_id]
    info_state = str(time_step.observations["info_state"][self._player_id])
    info_state_key = self._get_info_state_key(info_state)
    sampled_action, probs = None, None

    if not time_step.last():
      epsilon = 0.0 if is_evaluation else self._epsilon
      sampled_action, sample_policy, rm_policy = self._get_action_probs(
          info_state, legal_actions, epsilon
      )

    # Learn step: don't learn during evaluation or at first agent steps.
    if self._prev_sample_policy is not None and not is_evaluation:
      reward = time_step.rewards[self._player_id]
      values = np.zeros(self._num_actions, dtype=np.float64)
      values[self._prev_action] = (
          reward / self._prev_sample_policy[self._prev_action]
      )
      exp_value = np.dot(values, self._prev_rm_policy)
      for action in legal_actions:
        self._regrets[self._prev_info_state_key][action] += (
            values[action] - exp_value
        )

      if time_step.last():  # prepare for the next episode.
        self._prev_sample_policy = None
        return

    if not is_evaluation:
      self._prev_info_state_key = info_state_key
      self._prev_action = sampled_action
      self._prev_sample_policy = sample_policy
      self._prev_rm_policy = rm_policy
      self._prev_legal_actions = legal_actions
      self._prev_info_state = info_state

    return rl_agent.StepOutput(action=sampled_action, probs=probs)


class BotAgent(rl_agent.AbstractAgent):
  """Agent class that wraps a bot.

  Note, the environment must include the OpenSpiel state.
  """

  def __init__(self, num_actions, bot, name="bot_agent"):
    assert num_actions > 0
    self._bot = bot
    self._num_actions = num_actions

  def restart(self):
    self._bot.restart()

  def step(self, time_step, is_evaluation=False):
    # If it is the end of the episode, don't select an action.
    if time_step.last():
      return

    _, state = pyspiel.deserialize_game_and_state(
        time_step.observations["serialized_state"]
    )

    action = self._bot.step(state)
    probs = np.zeros(self._num_actions)
    probs[action] = 1.0

    return rl_agent.StepOutput(action=action, probs=probs)


def create_training_agent(
    agent_type,
    num_actions,
    info_state_size,
    hidden_layers_sizes,
    max_abs_reward,
    rng_seed,
    player_id,
):
  """Create training agent."""
  if agent_type == "dqn":
    return dqn.DQN(
        player_id=player_id,
        state_representation_size=info_state_size,
        num_actions=num_actions,
        discount_factor=0.99,
        epsilon_start=1.0,
        epsilon_end=0.1,
        hidden_layers_sizes=hidden_layers_sizes,
        learning_rate=FLAGS.learning_rate,
        replay_buffer_capacity=FLAGS.replay_buffer_capacity,
        batch_size=FLAGS.batch_size,
    )
  elif agent_type == "bdqn":
    return boltzmann_dqn.BoltzmannDQN(
        player_id=player_id,
        state_representation_size=info_state_size,
        num_actions=num_actions,
        discount_factor=0.99,
        epsilon_start=1.0,
        epsilon_end=0.1,
        hidden_layers_sizes=hidden_layers_sizes,
        learning_rate=FLAGS.learning_rate,
        replay_buffer_capacity=FLAGS.replay_buffer_capacity,
        batch_size=FLAGS.batch_size,
        eta=FLAGS.eta,
        seed=FLAGS.seed,
    )
  elif agent_type == "qlearning":
    return tabular_qlearner.QLearner(
        player_id=player_id,
        num_actions=num_actions,
        step_size=FLAGS.learning_rate,
        epsilon_schedule=rl_tools.LinearSchedule(0.5, 0.2, 1000000),
        discount_factor=0.99,
    )
  elif agent_type == "a2c":
    return policy_gradient.PolicyGradient(
        player_id,
        info_state_size,
        num_actions,
        loss_str="a2c",
        critic_learning_rate=FLAGS.critic_learning_rate,
        pi_learning_rate=FLAGS.pi_learning_rate,
        entropy_cost=FLAGS.entropy_cost,
        num_critic_before_pi=FLAGS.num_critic_before_pi,
        lambda_=FLAGS.lambda_,
        additional_discount_factor=0.99,
        hidden_layers_sizes=hidden_layers_sizes,
    )
  elif agent_type == "impala":
    return impala.IMPALA(  # pylint: disable=g-complex-comprehension
        player_id=player_id,
        state_representation_size=info_state_size,
        num_actions=num_actions,
        num_players=2,
        unroll_len=FLAGS.unroll_length,
        net_factory=impala.BasicRNN,
        rng_key=jax.random.PRNGKey(rng_seed),
        max_abs_reward=max_abs_reward,
        learning_rate=FLAGS.pi_learning_rate,
        entropy=FLAGS.entropy_cost,
        hidden_layers_sizes=hidden_layers_sizes,
        num_predictions=pyspiel.ROSHAMBO_NUM_BOTS + 1,
        prediction_weight=FLAGS.prediction_weight,
        batch_size=FLAGS.batch_size,
    )
  elif agent_type == "rm":
    return RegretMatchingAgent(
        player_id=player_id, num_actions=num_actions, epsilon=FLAGS.rm_epsilon
    )
  elif agent_type == "rock":
    return ConstantActionAgent(player_id, num_actions, 0)
  elif agent_type == "paper":
    return ConstantActionAgent(player_id, num_actions, 1)
  elif agent_type == "scissors":
    return ConstantActionAgent(player_id, num_actions, 2)
  elif agent_type == "uniform":
    return random_agent.RandomAgent(player_id, num_actions)
  else:
    assert False


def sample_bot_agent(pid, bot_names, population_ids, num_actions):
  idx = np.random.randint(0, len(population_ids))
  bot_id = population_ids[idx]
  name = bot_names[bot_id]
  bot = pyspiel.make_roshambo_bot(pid, name)
  return BotAgent(num_actions, bot, name=name), bot_id


class RollingAverage(object):
  """Class to store a rolling average."""

  def __init__(self, size=100):
    self._size = size
    self._values = np.array([0] * self._size, dtype=np.float64)
    self._index = 0
    self._total_additions = 0

  def add(self, value):
    self._values[self._index] = value
    self._total_additions += 1
    self._index = (self._index + 1) % self._size

  def mean(self):
    n = min(self._size, self._total_additions)
    if n == 0:
      return 0
    return self._values.sum() / n


def train_test_split(roshambo_bot_ids):
  """Create a train/test split for the roshambo bots."""

  if FIXED_POPULATION is not None:
    training_ids = FIXED_POPULATION[:]
    testing_ids = FIXED_POPULATION[:]
  elif FLAGS.pop_only >= 0:
    # If the pop_only flag is set, make a population of just that member
    assert FLAGS.pop_only < len(roshambo_bot_ids)
    training_ids = [FLAGS.pop_only]
    testing_ids = [FLAGS.pop_only]
  else:
    # Otherwise, do the train/test split
    bot_ids_copy = roshambo_bot_ids.copy()
    training_ids = list(bot_ids_copy.values())
    testing_ids = []
    if FLAGS.leave_out_set_size == 0:
      testing_ids = training_ids[:]
    else:
      while len(testing_ids) < FLAGS.leave_out_set_size:
        idx = np.random.randint(0, len(training_ids))
        testing_ids.append(training_ids[idx])
        training_ids.pop(idx)
  return training_ids, testing_ids


def print_roshambo_bot_names_and_ids(roshambo_bot_names):
  for i, name in enumerate(roshambo_bot_names):
    print(f"{i}: {name}")


class AgentBot(pyspiel.Bot):
  """An agent that wraps a bot."""

  def __init__(self, agent):
    pyspiel.Bot.__init__(self)
    self._agent = agent
    self._env = rl_environment.Environment(
        "repeated_game(stage_game=matrix_rps(),num_repetitions="
        + f"{pyspiel.ROSHAMBO_NUM_THROWS},"
        + f"recall={FLAGS.env_recall})",
        include_full_state=True,
    )

  def step(self, state):
    self._env.set_state(state)
    time_step = self._env.get_time_step()
    agent_output = self._agent.step(time_step, is_evaluation=True)
    return agent_output.action


def eval_checkpoint(roshambo_bot_names, prediction_logger):
  """Evaluate a checkpoint."""

  print("Starting eval checkpoint")
  print("Loading checkpoint")
  checkpoint = Checkpoint(FLAGS.eval_checkpoint)
  checkpoint.restore_or_save()
  assert checkpoint.state.learning_agents is not None
  print("Checkpoint loaded")
  greenberg_bot = pyspiel.make_roshambo_bot(1, "greenberg")
  greenberg_agent = BotAgent(3, greenberg_bot, name="greenberg_agent")
  print("Starting eval for agent...")
  env = rl_environment.Environment(
      "repeated_game(stage_game=matrix_rps(),num_repetitions="
      + f"{pyspiel.ROSHAMBO_NUM_THROWS},"
      + f"recall={FLAGS.env_recall})",
      include_full_state=True,
  )
  sum_eval_returns = np.zeros(pyspiel.ROSHAMBO_NUM_BOTS)
  for j in range(50):
    print(f"Eval checkpoint, j={j}")
    _, pop_expl = eval_agent(
        env,
        2,
        3,
        roshambo_bot_names,
        # checkpoint.state.learning_agents[1],
        greenberg_agent,
        prediction_logger,
        0,
    )
    eval_returns = (-1) * pop_expl
    sum_eval_returns += eval_returns
    avg_eval_returns = sum_eval_returns / (j + 1)
    pop_return = avg_eval_returns.sum() / pyspiel.ROSHAMBO_NUM_BOTS
    wp_expl = avg_eval_returns.min() * (-1)
    print(f"Pop return: {pop_return}, WP expl: {wp_expl}")
    print(avg_eval_returns)


def main(_):
  np.random.seed(FLAGS.seed)

  envs = [None, None]
  envs[0] = rl_environment.Environment(
      "repeated_game(stage_game=matrix_rps(),num_repetitions="
      + f"{pyspiel.ROSHAMBO_NUM_THROWS},"
      + f"recall={FLAGS.env_recall})",
      include_full_state=True,
  )
  envs[1] = rl_environment.Environment(
      "repeated_game(stage_game=matrix_rps(),num_repetitions="
      + f"{pyspiel.ROSHAMBO_NUM_THROWS},"
      + f"recall={FLAGS.env_recall})",
      include_full_state=True,
  )
  num_players = 2
  max_abs_reward = max(
      abs(envs[0].game.min_utility()), abs(envs[0].game.max_utility())
  )

  info_state_size = envs[0].observation_spec()["info_state"][0]
  num_actions = envs[0].action_spec()["num_actions"]

  print("Loading population...")
  pop_size = pyspiel.ROSHAMBO_NUM_BOTS
  print(f"Population size: {pop_size}")
  roshambo_bot_names = pyspiel.roshambo_bot_names()
  roshambo_bot_names.sort()
  print_roshambo_bot_names_and_ids(roshambo_bot_names)

  bot_id = 0
  roshambo_bot_ids = {}
  for name in roshambo_bot_names:
    roshambo_bot_ids[name] = bot_id
    bot_id += 1

  print(f"Leave out set size: {FLAGS.leave_out_set_size}")
  train_pop_ids, test_pop_ids = train_test_split(roshambo_bot_ids)
  print(f"Training ids: {train_pop_ids}")
  print(f"Test pop ids: {test_pop_ids}")

  if FLAGS.eval_checkpoint is not None:
    prediction_logger = PredictionLogger(FLAGS.pred_logs_dir)
    eval_checkpoint(roshambo_bot_names, prediction_logger)
    return

  rolling_averager = RollingAverage(FLAGS.window_size)
  expl_rolling_averagers = []
  for _ in range(pyspiel.ROSHAMBO_NUM_BOTS):
    expl_rolling_averagers.append(RollingAverage(FLAGS.window_size))

  print("Looking for checkpoint.")
  if FLAGS.cp_dir is None:
    print("cp_dir is None, disabling checkpointing.")
    # checkpoint = phoenix.Checkpoint()
    checkpoint = None
  else:
    print(f"Looking for checkpoint in {FLAGS.cp_dir}")
    checkpoint = Checkpoint(FLAGS.cp_dir)
    checkpoint.restore_or_save()
    print(f"Checkpoint loaded. ep = {checkpoint.state.ep}")

  if FLAGS.interactive_mode is not None:
    # Must restore an agent from a checkpoint
    assert checkpoint.state.ep is not None
    assert checkpoint.state.learning_agents is not None
    interactive_episode(
        envs[0],
        num_players,
        num_actions,
        roshambo_bot_names,
        checkpoint.state.learning_agent,
    )

  ep = None
  if checkpoint is not None:
    ep = checkpoint.state.ep
    if checkpoint.state.rolling_averager is not None:
      rolling_averager = checkpoint.state.rolling_averager
    if checkpoint.state.expl_rolling_averagers is not None:
      expl_rolling_averagers = checkpoint.state.expl_rolling_averagers
    if checkpoint.state.np_rng_state is not None:
      print("Restoring numpy random state")
      np.random.set_state(checkpoint.state.np_rng_state)

  if ep is None:
    ep = 0
  prediction_logger = PredictionLogger(FLAGS.pred_logs_dir)
  if FLAGS.pred_logs_dir is not None:
    pass  # TODO(author5): Add back in (make full director)

  hidden_layers_sizes = [int(l) for l in FLAGS.hidden_layers_sizes]
  # pylint: disable=g-complex-comprehension
  if checkpoint is None or checkpoint.state.learning_agents is None:
    learning_agents = [
        create_training_agent(
            FLAGS.learner,
            num_actions,
            info_state_size,
            hidden_layers_sizes,
            max_abs_reward,
            np.random.randint(100000000),
            player_id,
        )
        for player_id in [0, 1]
    ]
  else:
    learning_agents = checkpoint.state.learning_agents

  print(f"Starting at ep {ep}.")
  total_train_time = 0

  print("Starting training loop...")
  while ep < FLAGS.num_train_episodes:
    # Checkpoint save.
    if checkpoint is not None and ep > 0 and ep % FLAGS.cp_freq == 0:
      print("")
      print(f"Saving checkpoint at ep {ep}...")
      checkpoint.state.ep = ep
      checkpoint.state.np_rng_state = np.random.get_state()
      checkpoint.state.learning_agents = learning_agents
      checkpoint.state.rolling_averager = rolling_averager
      checkpoint.state.expl_rolling_averagers = expl_rolling_averagers
      checkpoint.save()
      print("Done saving checkpoint.")

    if (ep + 1) % FLAGS.eval_every == 0:
      print("")
      eps_per_sec = (ep + 1) / total_train_time
      print(f"Starting eval at ep {ep}. Avg train eps per sec: {eps_per_sec}")
      start_time_eval = time.time()
      eval_returns, pop_expl = eval_agent(
          envs[0],
          num_players,
          num_actions,
          roshambo_bot_names,
          learning_agents[1],
          prediction_logger,
          ep + 1,
      )
      value = eval_returns[1]
      rolling_averager.add(value)
      max_pop_exp = -1000
      for i in range(pyspiel.ROSHAMBO_NUM_BOTS):
        expl_rolling_averagers[i].add(pop_expl[i])
        max_pop_exp = max(max_pop_exp, expl_rolling_averagers[i].mean())
      r_mean = rolling_averager.mean()
      end_time_eval = time.time()
      print(f"Time for eval: {end_time_eval - start_time_eval}")
      data = {
          "episodes": ep + 1,
          "value": value,
          "swa_value": r_mean,
          "expl_swa_value": max_pop_exp,
          "agg_score_swa": r_mean - max_pop_exp,
          "eps_per_sec": eps_per_sec,
      }
      print(data)
      sys.stdout.flush()

    ep_start_time = time.time()
    for learner_pid in range(2):
      agents = [None, None]
      agents[learner_pid] = learning_agents[learner_pid]
      env = envs[learner_pid]
      assert env is not None
      # print(f"Learner pid: {learner_pid}")
      roll = np.random.uniform()

      if roll < FLAGS.prob_selfplay:
        agents[1 - learner_pid] = learning_agents[1 - learner_pid]
        env.set_prediction_label(pyspiel.ROSHAMBO_NUM_BOTS)
      else:
        pop_agent, pop_idx = sample_bot_agent(
            1 - learner_pid, roshambo_bot_names, train_pop_ids, num_actions
        )
        agents[1 - learner_pid] = pop_agent
        env.set_prediction_label(pop_idx)

      time_step = env.reset()
      while not time_step.last():
        time_step2 = copy.deepcopy(time_step)
        player_id = time_step.observations["current_player"]
        agents_output = [agents[0].step(time_step), agents[1].step(time_step2)]
        action_list = [agent_output.action for agent_output in agents_output]
        time_step = env.step(action_list)

      # Episode is over, step all agents with final info state.
      time_step2 = copy.deepcopy(time_step)
      assert agents[0] is not None
      assert agents[1] is not None
      agents[0].step(time_step)
      agents[1].step(time_step2)

    ep_end_time = time.time()
    total_train_time += ep_end_time - ep_start_time
    ep += 1
    print(".", end="")
    sys.stdout.flush()


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Stochastic Gradient Descent (SGD) Approx. Nash Solver."""

from absl import logging  # pylint:disable=unused-import

import numpy as np

from open_spiel.python.algorithms.adidas_utils.helpers import simplex
from open_spiel.python.algorithms.adidas_utils.helpers.symmetric import exploitability as exp
from open_spiel.python.algorithms.adidas_utils.helpers.symmetric import updates


class Solver(updates.Solver):
  """SGD Solver."""

  def __init__(self, temperature=0., proj_grad=True, euclidean=False,
               lrs=(1e-1,), rnd_init=False, seed=None, **kwargs):
    """Ctor."""
    del kwargs
    super().__init__(proj_grad, euclidean, rnd_init, seed)
    if temperature < 0.:
      raise ValueError('temperature must be non-negative')
    self.temperature = temperature
    self.lrs = lrs
    self.num_estimates = 2

    if temperature > 0:
      self.eps = np.exp(-1 / temperature)  # ensure dist[i] >= eps / dim(dist)
    else:
      self.eps = 0.
    if euclidean:
      self.update = lambda *args: self.euc_descent_step(*args, eps=self.eps)
    else:
      self.update = lambda *args: self.mirror_descent_step(*args, eps=self.eps)

  def init_vars(self, num_strats, num_players):
    """Initialize solver parameters."""
    self.num_players = num_players
    if self.rnd_init:
      init_dist = self.random.rand(num_strats)
    else:
      init_dist = np.ones(num_strats)
    init_dist /= init_dist.sum()
    init_dist = simplex.project_to_interior(init_dist, self.eps)
    return (init_dist,)

  def compute_gradients(self, params, payoff_matrices):
    """Compute and return exploitability.

    Args:
      params: tuple of params (dist,), see sgd.gradients
      payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    Returns:
      float, exploitability of current dist
      unregularized exploitability (stochastic estimate)
      shannon regularized exploitability (stochastic estimate)
    """
    return gradients(*params, payoff_matrices, self.num_players,
                     self.temperature, self.proj_grad)

  def exploitability(self, params, payoff_matrices):
    """Compute and return exploitability.

    Args:
      params: tuple of params (dist,), see sgd.gradients
      payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    Returns:
      float, exploitability as avg squared norm of projected-gradient
    """
    return exp.grad_norm_exploitability(params, payoff_matrices, eta=1.,
                                        temperature=self.temperature)


def gradients(dist, payoff_matrices, num_players, temperature=0.,
              proj_grad=True):
  """Computes exploitablity gradient.

  Assumption: eta_k = 1 for all k

  Args:
    dist: 1-d np.array, current estimate of nash distribution
    payoff_matrices: 2 (>=2 x A x A) np.arrays, payoffs for each joint action
    num_players: int, number of players, in case payoff_matrices is abbreviated
    temperature: non-negative float, default 0.
    proj_grad: bool, if True, projects dist gradient onto simplex
  Returns:
    gradient of exploitability w.r.t. (dist) as tuple
    unregularized exploitability (stochastic estimate)
    shannon regularized exploitability (stochastic estimate)
  """
  del num_players
  tau = temperature

  a, b = 0, 1  # 2 samples needed for unbiased estimation
  p_0, p_1 = 0, 1  # player 0 index, player 1 index
  hess_0_01_a = payoff_matrices[a][p_0]
  hess_1_01_a = payoff_matrices[a][p_1]
  hess_0_01_b = payoff_matrices[b][p_0]

  pg_0_a = simplex.project_grad(hess_0_01_a.dot(dist))
  pg_0_b = simplex.project_grad(hess_0_01_b.dot(dist))

  unreg_exp = np.dot(pg_0_a, pg_0_b)

  if tau > 0.:
    log_dist_safe = np.clip(np.log(dist), -40, 0)
    entr_grad_proj = simplex.project_grad(-tau * (log_dist_safe + 1))
  else:
    entr_grad_proj = 0.
  pg_0_a_entr = pg_0_a + entr_grad_proj
  pg_0_b_entr = pg_0_b + entr_grad_proj
  pg_0_entr = 0.5 * (pg_0_a_entr + pg_0_b_entr)
  pg_1_b_entr = pg_0_b_entr

  reg_exp = np.dot(pg_0_a_entr, pg_0_b_entr)

  # then construct unbiased stochastic gradient
  grad_dist = 2. * hess_1_01_a.dot(pg_1_b_entr)
  if tau > 0.:
    grad_dist += 2. * -tau * pg_0_entr / dist

  if proj_grad:
    grad_dist = simplex.project_grad(grad_dist)

  return (grad_dist,), unreg_exp, reg_exp



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""JAX implementation of the counterfactual regret minimization algorithm usable with GPU acceleration.

Uses same CFR setting as open_spiel.python.algorithms.cfr._CFRSolverBase and the
usability should be interchangable.

The results may slightly differ between these 2 versions due to rounding errors
when computing regrets (rounding regrets smaller than epsilon to zero results in
exactly the same results)

The algorithm performs well in short but wide games, with small amount of
illegal actions and poorly in long games with a lot of illegal actions.
"""

# pylint: disable=g-importing-member

from collections import namedtuple
import functools

import chex
import jax
import jax.numpy as jnp
import numpy as np

from open_spiel.python import policy
import pyspiel

JAX_CFR_SIMULTANEOUS_UPDATE = -5


def regret_matching(regret, mask):
  """Computes current policy based on current regrets.

  Args:
    regret: Current regrets in array Fkiat[Isets, Actions]
    mask: Legal action mask Bool[Isets, Actions]

  Returns:
    policy: the policy.
  """
  regret = jnp.maximum(regret, 0) * mask
  total = jnp.sum(regret, axis=-1, keepdims=True)

  return jnp.where(total > 0.0, regret / total, 1.0 / jnp.sum(mask)) * mask


def update_regrets_plus(regret):
  """Clamps the regrets to be non-negative."""
  return regret * (regret > 0)


def update_regrets(regret):
  """Updates the regrets without CFRPlus."""
  return regret


@chex.dataclass(frozen=True)
class JaxCFRConstants:
  """Constants for JaxCFR."""

  players: int
  max_depth: int
  # This includes chance outcomes!  TODO: We could do this separately for each
  # depth to make less computations.
  max_actions: int

  max_iset_depth: chex.ArrayTree = ()  # Is just a list of integers
  isets: chex.ArrayTree = ()  # Is just a list of integers

  depth_history_utility: chex.ArrayTree = ()
  depth_history_iset: chex.ArrayTree = ()
  depth_history_actions: chex.ArrayTree = ()
  depth_history_previous_iset: chex.ArrayTree = ()
  depth_history_previous_action: chex.ArrayTree = ()

  depth_history_next_history: chex.ArrayTree = ()
  depth_history_player: chex.ArrayTree = ()
  depth_history_chance: chex.ArrayTree = ()
  depth_history_previous_history: chex.ArrayTree = ()
  depth_history_action_mask: chex.ArrayTree = ()
  depth_history_chance_probabilities: chex.ArrayTree = ()

  iset_previous_action: chex.ArrayTree = ()
  iset_action_mask: chex.ArrayTree = ()
  iset_action_depth: chex.ArrayTree = ()


class JaxCFR:
  """Class for CFR and CFR.

  First it prepares all the structures in `init`, then it just reuses them
  within jitted function `jit_step`.
  """

  def __init__(
      self,
      game: pyspiel.Game,
      regret_matching_plus=True,
      alternating_updates=True,
      linear_averaging=True,
  ):
    self.game = game
    self._regret_matching_plus = regret_matching_plus
    self._alternating_updates = alternating_updates
    self._linear_averaging = linear_averaging
    self.timestep = 1

    self.init()

  def init(self):
    """Constructor."""

    players = self.game.num_players()
    depth_history_utility = [[] for _ in range(players)]
    depth_history_previous_iset = [[] for _ in range(players)]
    depth_history_previous_action = [[] for _ in range(players)]
    depth_history_iset = [[] for _ in range(players)]
    depth_history_actions = [[] for _ in range(players)]
    depth_history_next_history = []
    depth_history_player = []
    depth_history_chance = []
    depth_history_previous_history = []
    depth_history_action_mask = []
    depth_history_chance_probabilities = []
    # Previous action is mapping of both iset and action!
    iset_previous_action = [[] for _ in range(players)]
    iset_action_mask = [[] for _ in range(players)]
    iset_action_depth = [[] for _ in range(players)]
    ids = [0 for _ in range(players)]
    pl_isets = [{} for _ in range(players)]
    distinct_actions = max(
        self.game.num_distinct_actions(), self.game.max_chance_outcomes()
    )

    for pl in range(players):
      pl_isets[pl][''] = ids[pl]
      ids[pl] += 1
      am = [0] * distinct_actions
      am[0] = 1
      iset_action_mask[pl].append(am)
      iset_previous_action[pl].append(0)
      iset_action_depth[pl].append(0)

    PreviousInfo = namedtuple(
        'PreviousInfo',
        ('actions', 'isets', 'prev_actions', 'history', 'player'),
    )

    def _traverse_tree(state, previous_info, depth, chance=1.0):

      if len(depth_history_next_history) <= depth:
        for pl in range(players):
          depth_history_utility[pl].append([])
          depth_history_previous_iset[pl].append([])
          depth_history_previous_action[pl].append([])
          depth_history_iset[pl].append([])
          depth_history_actions[pl].append([])

        depth_history_next_history.append([])
        depth_history_player.append([])
        depth_history_chance.append([])
        depth_history_previous_history.append([])
        depth_history_action_mask.append([])
        depth_history_chance_probabilities.append([])

      history_id = len(depth_history_previous_history[depth])

      next_history_temp = [0] * distinct_actions
      depth_history_next_history[depth].append(next_history_temp)
      depth_history_player[depth].append(state.current_player())
      depth_history_chance[depth].append(chance)
      depth_history_previous_history[depth].append(previous_info.history)

      actions_mask = [0] * distinct_actions
      for a in state.legal_actions():
        actions_mask[a] = 1
      depth_history_action_mask[depth].append(actions_mask)
      chance_probabilities = [0.0 for _ in range(distinct_actions)]
      if state.is_chance_node():
        for a, prob in state.chance_outcomes():
          chance_probabilities[a] = prob
      elif not state.is_terminal():
        chance_probabilities = [1.0 for _ in range(distinct_actions)]
      else:
        chance_probabilities = [
            1.0 / distinct_actions for _ in range(distinct_actions)
        ]

      depth_history_chance_probabilities[depth].append(chance_probabilities)
      for pl in range(players):
        depth_history_utility[pl][depth].append(
            state.rewards()[pl] if not state.is_chance_node() else 0.0
        )
        depth_history_previous_iset[pl][depth].append(previous_info.isets[pl])
        depth_history_previous_action[pl][depth].append(
            previous_info.actions[pl]
        )
        if state.current_player() == pl:
          iset = state.information_state_string()
          if iset not in pl_isets[pl]:
            pl_isets[pl][iset] = ids[pl]
            ids[pl] += 1
            iset_previous_action[pl].append(previous_info.actions[pl])
            iset_action_mask[pl].append(actions_mask)
            iset_action_depth[pl].append(previous_info.prev_actions[pl])
          depth_history_iset[pl][depth].append(pl_isets[pl][iset])
          depth_history_actions[pl][depth].append([
              i + pl_isets[pl][iset] * distinct_actions
              for i in range(distinct_actions)
          ])
        else:
          depth_history_iset[pl][depth].append(0)
          depth_history_actions[pl][depth].append(
              [0 for _ in range(distinct_actions)]
          )

      for a in state.legal_actions():
        new_chance = chance * chance_probabilities[a]
        assert new_chance > 0.0
        new_actions = tuple(
            previous_info.actions[pl]
            if state.current_player() != pl
            else pl_isets[pl][iset] * distinct_actions + a
            for pl in range(players)
        )
        new_infosets = tuple(
            previous_info.isets[pl]
            if state.current_player() != pl
            else pl_isets[pl][iset]
            for pl in range(players)
        )
        new_prev_actions = tuple(
            previous_info.prev_actions[pl] + int(state.current_player() == pl)
            for pl in range(players)
        )
        new_info = PreviousInfo(
            new_actions,
            new_infosets,
            new_prev_actions,
            history_id,
            state.current_player(),
        )
        new_state = state.clone()
        new_state.apply_action(a)

        # simple workaround if the next element was not visited yet
        next_history_temp[a] = (
            len(depth_history_player[depth + 1])
            if len(depth_history_player) > depth + 1
            else 0
        )

        _traverse_tree(new_state, new_info, depth + 1, new_chance)

    s = self.game.new_initial_state()
    _traverse_tree(
        s,
        PreviousInfo(
            tuple(0 for _ in range(players)),
            tuple(0 for _ in range(players)),
            tuple(0 for _ in range(players)),
            0,
            0,
        ),
        0,
    )

    def convert_to_jax(x):
      return [jnp.asarray(i) for i in x]

    def convert_to_jax_players(x):
      return [[jnp.asarray(i) for i in x[pl]] for pl in range(players)]

    depth_history_utility = convert_to_jax_players(depth_history_utility)
    depth_history_iset = convert_to_jax_players(depth_history_iset)
    depth_history_previous_iset = convert_to_jax_players(
        depth_history_previous_iset
    )
    depth_history_actions = convert_to_jax_players(depth_history_actions)
    depth_history_previous_action = convert_to_jax_players(
        depth_history_previous_action
    )

    depth_history_next_history = convert_to_jax(depth_history_next_history)
    depth_history_player = convert_to_jax(depth_history_player)
    depth_history_chance = convert_to_jax(depth_history_chance)
    depth_history_previous_history = convert_to_jax(
        depth_history_previous_history
    )
    depth_history_chance_probabilities = convert_to_jax(
        depth_history_chance_probabilities
    )
    depth_history_action_mask = convert_to_jax(depth_history_action_mask)

    max_iset_depth = [np.max(iset_action_depth[pl]) for pl in range(players)]
    iset_previous_action = convert_to_jax(iset_previous_action)
    iset_action_mask = convert_to_jax(iset_action_mask)
    iset_action_depth = convert_to_jax(iset_action_depth)

    self.constants = JaxCFRConstants(
        players=players,
        max_depth=int(len(depth_history_utility[0])),
        max_actions=distinct_actions,
        max_iset_depth=max_iset_depth,
        isets=ids,
        depth_history_utility=depth_history_utility,
        depth_history_iset=depth_history_iset,
        depth_history_actions=depth_history_actions,
        depth_history_previous_iset=depth_history_previous_iset,
        depth_history_previous_action=depth_history_previous_action,
        depth_history_next_history=depth_history_next_history,
        depth_history_player=depth_history_player,
        depth_history_chance=depth_history_chance,
        depth_history_previous_history=depth_history_previous_history,
        depth_history_action_mask=depth_history_action_mask,
        depth_history_chance_probabilities=depth_history_chance_probabilities,
        iset_previous_action=iset_previous_action,
        iset_action_mask=iset_action_mask,
        iset_action_depth=iset_action_depth,
    )

    self.regrets = [
        jnp.zeros((ids[pl], distinct_actions)) for pl in range(players)
    ]
    self.averages = [
        jnp.zeros((ids[pl], distinct_actions)) for pl in range(players)
    ]

    self.regret_matching = jax.vmap(regret_matching, 0, 0)
    if self._regret_matching_plus:
      self.update_regrets = jax.vmap(update_regrets_plus, 0, 0)
    else:
      self.update_regrets = jax.vmap(update_regrets, 0, 0)

    self.iset_map = pl_isets

  def multiple_steps(self, iterations: int):
    """Performs several CFR steps.

    Args:
      iterations: Amount of CFR steps, the solver should do.
    """
    for _ in range(iterations):
      self.step()

  def evaluate_and_update_policy(self):
    """Wrapper to step().

    Ensures interchangability with
    open_spiel.python.algorithms.cfr._CFRSolverBase.
    """
    self.step()

  def step(self):
    """Wrapper around the jitted function for performing CFR step."""
    averaging_coefficient = self.timestep if self._linear_averaging else 1
    if self._alternating_updates:
      for player in range(self.constants.players):
        self.regrets, self.averages = self.jit_step(
            self.regrets, self.averages, averaging_coefficient, player
        )

    else:
      self.regrets, self.averages = self.jit_step(
          self.regrets,
          self.averages,
          averaging_coefficient,
          JAX_CFR_SIMULTANEOUS_UPDATE,
      )

    self.timestep += 1

  def propagate_strategy(self, current_strategies):
    """Propagtes the strategies withing infosets.

    Args:
      current_strategies: Current strategies for all players, list[Float[Isets,
        Actions]]
    Returns:
      realization_plans: the realization plans.
    """
    realization_plans = [
        jnp.ones_like(current_strategies[pl])
        for pl in range(self.constants.players)
    ]

    for pl in range(self.constants.players):
      for i in range(0, self.constants.max_iset_depth[pl] + 1):
        realization_plans[pl] = jnp.where(
            self.constants.iset_action_depth[pl][..., jnp.newaxis] == i,
            current_strategies[pl]
            * realization_plans[pl].ravel()[
                self.constants.iset_previous_action[pl]
            ][..., jnp.newaxis],
            realization_plans[pl],
        )

    return realization_plans

  @functools.partial(jax.jit, static_argnums=(0,))
  def jit_step(
      self, regrets, averages, average_policy_update_coefficient, player
  ):
    """Performs the CFR step.

    This consists of:
    1. Computes the current strategies based on regrets
    2. Computes the realization plan for each action from top of the tree down
    3. Compute the counterfactual regrets from bottom of the tree up
    4. Updates regrets and average stretegies

    Args:
      regrets: Cummulative regrets for all players, list[Float[Isets, Actions]]
      averages: Average strategies for all players, list[Float[Isets, Actions]]
      average_policy_update_coefficient: Weight of the average policy update.
        When enabled linear_averging it is equal to current iteration. Otherwise
        1, int
      player: Player for which the update should be done. When alternating
        updates are distables, it is JAX_CFR_SIMULTANEOUS_UPDATE

    Returns:
      regrets: the regrets.
      averages: the averages.
    """
    current_strategies = [
        self.regret_matching(regrets[pl], self.constants.iset_action_mask[pl])
        for pl in range(self.constants.players)
    ]

    realization_plans = self.propagate_strategy(current_strategies)
    iset_reaches = [
        jnp.sum(realization_plans[pl], -1)
        for pl in range(self.constants.players)
    ]
    # In last row, there are only terminal, so we start row before it
    depth_utils = [
        [self.constants.depth_history_utility[pl][-1]]
        for pl in range(self.constants.players)
    ]
    for i in range(self.constants.max_depth - 2, -1, -1):

      each_history_policy = self.constants.depth_history_chance_probabilities[i]
      for pl in range(self.constants.players):
        each_history_policy = each_history_policy * jnp.where(
            self.constants.depth_history_player[i][..., jnp.newaxis] == pl,
            current_strategies[pl][self.constants.depth_history_iset[pl][i]],
            1,
        )

      for pl in range(self.constants.players):
        action_value = jnp.where(
            self.constants.depth_history_player[i][..., jnp.newaxis] == -4,
            self.constants.depth_history_utility[pl][i][..., jnp.newaxis],
            depth_utils[pl][-1][self.constants.depth_history_next_history[i]],
        )
        history_value = jnp.sum(action_value * each_history_policy, -1)
        regret = (
            (action_value - history_value[..., jnp.newaxis])
            * self.constants.depth_history_action_mask[i]
            * (self.constants.depth_history_player[i][..., jnp.newaxis] == pl)
            * self.constants.depth_history_chance[i][..., jnp.newaxis]
        )
        for pl2 in range(self.constants.players):
          if pl != pl2:
            regret = (
                regret
                * realization_plans[pl2].ravel()[
                    self.constants.depth_history_previous_action[pl2][i]
                ][..., jnp.newaxis]
            )
        bin_regrets = jnp.bincount(
            self.constants.depth_history_actions[pl][i].ravel(),
            regret.ravel(),
            length=self.constants.isets[pl] * self.constants.max_actions,
        )
        bin_regrets = bin_regrets.reshape(-1, self.constants.max_actions)
        regrets[pl] = jnp.where(
            jnp.logical_or(player == pl, player == JAX_CFR_SIMULTANEOUS_UPDATE),
            regrets[pl] + bin_regrets,
            regrets[pl],
        )
        depth_utils[pl].append(history_value)

    regrets = [
        self.update_regrets(regrets[pl]) for pl in range(self.constants.players)
    ]

    averages = [
        jnp.where(
            jnp.logical_or(player == pl, player == JAX_CFR_SIMULTANEOUS_UPDATE),
            averages[pl]
            + current_strategies[pl]
            * iset_reaches[pl][..., jnp.newaxis]
            * average_policy_update_coefficient,
            averages[pl],
        )
        for pl in range(self.constants.players)
    ]

    return regrets, averages

  def average_policy(self):
    """Extracts the average policy from JAX structures into a TabularPolicy."""
    averages = [
        np.asarray(self.averages[pl]) for pl in range(self.constants.players)
    ]
    averages = [
        averages[pl] / np.sum(averages[pl], -1, keepdims=True)
        for pl in range(self.constants.players)
    ]

    avg_strategy = policy.TabularPolicy(self.game)

    for pl in range(2):
      for iset, val in self.iset_map[pl].items():
        if not iset:
          continue
        state_policy = avg_strategy.policy_for_key(iset)
        for i in range(len(state_policy)):
          state_policy[i] = averages[pl][val][i]
    return avg_strategy




# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An asynchronous Monte Carlo Tree Search algorithm for game play.

This implements asynchronous MCTS which runs evaluations asynchronously in
parallel. It is a very simplified version of the algorithm described in:

Schultz et al. '24. Mastering Board Games by External and Internal Planning
with Language Models. https://arxiv.org/abs/2412.12119

Unlike the paper which describes a entirely model-based approach, this
implementation does use a game engine internally. Also, this version uses
standard virtual losses (not the dynamic virtual counts described in the paper).
However, it does implement the asynchronous calls to the evaluators in the same
way as the algorithm in the paper.
"""

from __future__ import annotations
import concurrent.futures
import math
import time
from typing import Callable, Optional

from absl import logging
import numpy as np

import pyspiel


def robust_child_with_total_reward_tiebreaker(
    root: SearchNode,
) -> tuple[int, SearchNode]:
  """Returns the best action and associated child node.

  The child node with the most visits is chosen.
  In case of a tie, the child with the highest total reward is chosen.


  Args:
    root: The root node of the search tree.

  Returns:
    A tuple containing the best action and the associated child node.
  """
  selection_criteria = lambda node: (node.explore_count, node.total_reward)
  best_child = max(root.children, key=selection_criteria)
  return best_child.action, best_child


def robust_child(root: SearchNode) -> tuple[int, SearchNode]:
  """Returns the best action and associated child node.

  A child node with the most visits is chosen.

  Args:
    root: The root node of the search tree.

  Returns:
    A tuple containing the best action and the associated child node.
  """
  selection_criteria = lambda node: node.explore_count
  best_child = max(root.children, key=selection_criteria)
  return best_child.action, best_child


def max_child(root: SearchNode) -> tuple[int, SearchNode]:
  """Returns the best action and associated child node.

  A child node with the highest expected reward is chosen.

  Args:
    root: The root node of the search tree.

  Returns:
    A tuple containing the best action and the associated child node.
  """
  selection_criteria = (
      lambda node: node.total_reward / node.explore_count
      if node.explore_count
      else float("-inf")
  )
  best_child = max(root.children, key=selection_criteria)
  return best_child.action, best_child


def max_robust_child(
    root: SearchNode, find_robust: bool = False
) -> tuple[Optional[int], Optional[SearchNode]]:
  """Returns the best action and associated child node.

  A child node with the highest expected reward and most visits is chosen.
  If no such child exists, increase the number of simulations.

  Args:
    root: The root node of the search tree.
    find_robust: Whether to find a robust child node. E.g., if max compute is
      reached and max robust is not found.

  Returns:
    A tuple containing the best action and the associated child node.
  """
  if find_robust:
    best_action, best_child = robust_child(root)
  else:
    _, max_child_node = max_child(root)
    _, robust_child_node = robust_child(root)
    best_action, best_child = None, None
    for child in root.children:
      if child == max_child_node and child == robust_child_node:
        best_action, best_child = child.action, child
        break
  return best_action, best_child


def secure_child(
    root: SearchNode, secure_c: float = 1.0
) -> tuple[int, SearchNode]:
  """Returns the best action and associated child node.

  A child node with the most visits is chosen.

  Args:
    root: The root node of the search tree.
    secure_c: The constant used to calculate lower uncertainty bound

  Returns:
    A tuple containing the best action and the associated child node.
  """
  selection_criteria = (
      lambda node: node.total_reward / node.explore_count  # pylint: disable=g-long-ternary
      - secure_c / math.sqrt(node.explore_count)
      if node.explore_count
      else float("-inf")
  )
  best_child = max(root.children, key=selection_criteria)
  return best_child.action, best_child


def max_robust_secure_child(
    root: SearchNode, secure_c: float = 1.0, find_secure: bool = False
) -> tuple[Optional[int], Optional[SearchNode]]:
  """Returns the best action and associated child node.

  A child node with the most visits is chosen.

  Args:
    root: The root node of the search tree.
    secure_c: The constant used to calculate lower uncertainty bound.
    find_secure: Whether to find a secure child node.

  Returns:
    A tuple containing the best action and the associated child node.
  """
  if find_secure:
    best_action, best_child = secure_child(root, secure_c)
  else:
    best_action, best_child = max_robust_child(root)
  return best_action, best_child


class Evaluator:
  """Abstract class representing an evaluation function for a game.

  The evaluation function takes in an intermediate state in the game and returns
  an evaluation of that state, which should correlate with chances of winning
  the game. It returns the evaluation from all player's perspectives.
  """

  def prior_and_value(
      self, state: pyspiel.State
  ) -> tuple[list[tuple[int, float]], np.ndarray]:
    """Returrns a prior (list of (action, prior)) and values (np.array)."""
    raise NotImplementedError


class RandomRolloutEvaluator(Evaluator):
  """A simple evaluator doing random rollouts.

  This evaluator returns the average outcome of playing random actions from the
  given state until the end of the game.  n_rollouts is the number of random
  outcomes to be considered.
  """

  def __init__(
      self,
      random_state: np.random.RandomState | None = None,
  ):
    self._random_state = random_state or np.random.RandomState()

  def prior_and_value(
      self, state: pyspiel.State
  ) -> tuple[list[tuple[int, float]], np.ndarray]:
    """Returns evaluation on given state."""
    # prior
    if state.is_chance_node():
      prior = state.chance_outcomes()
    else:
      legal_actions = state.legal_actions(state.current_player())
      prior = [(action, 1.0 / len(legal_actions)) for action in legal_actions]
    # value
    working_state = state.clone()
    while not working_state.is_terminal():
      if working_state.is_chance_node():
        outcomes = working_state.chance_outcomes()
        action_list, prob_list = zip(*outcomes)
        action = self._random_state.choice(action_list, p=prob_list)
      else:
        action = self._random_state.choice(working_state.legal_actions())
      working_state.apply_action(action)
    value = np.array(working_state.returns())
    return prior, value


class SearchNode(object):
  """A node in the search tree.

  A SearchNode represents a state and possible continuations from it. Each child
  represents a possible action, and the expected result from doing so.

  Attributes:
    action: The action from the parent node's perspective. Not important for the
      root node, as the actions that lead to it are in the past.
    player: Which player made this action.
    prior: A prior probability for how likely this action will be selected.
    explore_count: How many times this node was explored.
    total_reward: The sum of rewards of rollouts through this node, from the
      parent node's perspective. The average reward of this node is
      `total_reward / explore_count`
    outcome: The rewards for all players if this is a terminal node or the
      subtree has been proven, otherwise None.
    children: A list of SearchNodes representing the possible actions from this
      node, along with their expected rewards.
    expanded: Whether this node has been expanded.
  """

  __slots__ = [
      "action",
      "player",
      "prior",
      "explore_count",
      "total_reward",
      "outcome",
      "children",
      "expanded",
  ]

  def __init__(self, action: int | None, player: int, prior: float):
    self.action = action
    self.player = player
    self.prior = prior
    self.explore_count = 0
    self.total_reward = 0.0
    self.outcome = None
    self.children = []
    self.expanded = False

  def uct_value(self, parent_explore_count: int, uct_c: float) -> float:
    """Returns the UCT value of child."""
    if self.outcome is not None:
      return self.outcome[self.player]

    if self.explore_count == 0:
      return float("inf")

    return self.total_reward / self.explore_count + uct_c * math.sqrt(
        math.log(parent_explore_count) / self.explore_count
    )

  def puct_value(self, parent_explore_count: int, uct_c: float) -> float:
    """Returns the PUCT value of child."""
    if self.outcome is not None:
      return self.outcome[self.player]

    return (
        self.explore_count and self.total_reward / self.explore_count
    ) + uct_c * self.prior * math.sqrt(parent_explore_count) / (
        self.explore_count + 1
    )

  def sort_key(self):
    """Returns the best action from this node, either proven or most visited.

    This ordering leads to choosing:
    - Highest proven score > 0 over anything else, including a promising but
      unproven action.
    - A proven draw only if it has higher exploration than others that are
      uncertain, or the others are losses.
    - Uncertain action with most exploration over loss of any difficulty
    - Hardest loss if everything is a loss
    - Highest expected reward if explore counts are equal (unlikely).
    - Longest win, if multiple are proven (unlikely due to early stopping).
    """
    return (
        0 if self.outcome is None else self.outcome[self.player],
        self.explore_count,
        self.total_reward,
    )

  def best_child(self):
    """Returns the best child in order of the sort key."""
    return max(self.children, key=SearchNode.sort_key)

  def children_str(self, state=None):
    """Returns the string representation of this node's children.

    They are ordered based on the sort key, so order of being chosen to play.

    Args:
      state: A `pyspiel.State` object, to be used to convert the action id into
        a human readable format. If None, the action integer id is used.
    """
    return "\n".join([
        c.to_str(state)
        for c in reversed(sorted(self.children, key=SearchNode.sort_key))
    ])

  def to_str(self, state=None):
    """Returns the string representation of this node.

    Args:
      state: A `pyspiel.State` object, to be used to convert the action id into
        a human readable format. If None, the action integer id is used.
    """
    action = (
        state.action_to_string(state.current_player(), self.action)
        if state and self.action is not None
        else str(self.action)
    )
    return (
        "{:>6}: player: {}, prior: {:5.3f}, value: {:6.3f}, sims: {:5d}, "
        "outcome: {}, {:3d} children"
    ).format(
        action,
        self.player,
        self.prior,
        self.explore_count and self.total_reward / self.explore_count,
        self.explore_count,
        (
            "{:4.1f}".format(self.outcome[self.player])
            if self.outcome
            else "none"
        ),
        len(self.children),
    )

  def __str__(self):
    return self.to_str(None)


class MCTSBot(pyspiel.Bot):
  """Bot that uses Monte-Carlo Tree Search algorithm."""

  def __init__(
      self,
      game,
      uct_c,
      max_simulations,
      evaluator,
      solve=True,
      random_state=None,
      child_selection_fn=SearchNode.uct_value,
      best_child_fn: Callable[
          ..., tuple[Optional[int], Optional[SearchNode]]
      ] = robust_child_with_total_reward_tiebreaker,
      dirichlet_noise=None,
      verbose=False,
      dont_return_chance_node=False,
      virtual_loss: int = 10,  # virtual loss for async MCTS
      batch_size: int = 16,  # batch size for asynchronous MCTS
      secure_c: float = 1.0,  # secure constant for secure child selection
      simulations_multiplier: float = 1.0,
      max_additional_simulation_rounds: int = 0,
      timeout: float = 5.0,  # timeout for asynchronous MCTS
  ):
    """Initializes a MCTS Search algorithm in the form of a bot.

    In multiplayer games, or non-zero-sum games, the players will play the
    greedy strategy.

    Args:
      game: A pyspiel.Game to play.
      uct_c: The exploration constant for UCT.
      max_simulations: How many iterations of MCTS to perform. Each simulation
        will result in one call to the evaluator. Memory usage should grow
        linearly with simulations * branching factor. How many nodes in the
        search tree should be evaluated. This is correlated with memory size and
        tree depth.
      evaluator: A `Evaluator` object to use to evaluate a leaf node.
      solve: Whether to back up solved states.
      random_state: An optional numpy RandomState to make it deterministic.
      child_selection_fn: A function to select the child in the descent phase.
        The default is UCT.
      best_child_fn: A function to select the best child in root node after tree
        is built. The default is the child with the most visits.
      dirichlet_noise: A tuple of (epsilon, alpha) for adding dirichlet noise to
        the policy at the root. This is from the alpha-zero paper.
      verbose: Whether to print information about the search tree before
        returning the action. Useful for confirming the search is working
        sensibly.
      dont_return_chance_node: If true, do not stop expanding at chance nodes.
        Enabled for AlphaZero.
      virtual_loss: the value to use for virtual loss in async MCTS.
      batch_size: The batch size for asynchronous MCTS.
      secure_c: The constant used to calculate lower uncertainty bound
      simulations_multiplier: The multiplier for search budget.
      max_additional_simulation_rounds: The maximum number of additional
        simulation rounds.
      timeout: The timeout for asynchronous MCTS.

    Raises:
      ValueError: if the game type isn't supported.
    """
    pyspiel.Bot.__init__(self)
    # Check that the game satisfies the conditions for this MCTS implementation.
    game_type = game.get_type()
    if game_type.reward_model != pyspiel.GameType.RewardModel.TERMINAL:
      raise ValueError("Game must have terminal rewards.")
    if game_type.dynamics != pyspiel.GameType.Dynamics.SEQUENTIAL:
      raise ValueError("Game must have sequential turns.")

    self._game = game
    self.uct_c = uct_c
    self.max_simulations = max_simulations
    self._max_simulations = max_simulations
    self.evaluator = evaluator
    self.verbose = verbose
    self.solve = solve
    self.max_utility = game.max_utility()
    self.min_utility = game.min_utility()
    self._dirichlet_noise = dirichlet_noise
    self._random_state = random_state or np.random.RandomState()
    self._child_selection_fn = child_selection_fn
    self._best_child_fn = best_child_fn
    self.dont_return_chance_node = dont_return_chance_node
    self._root = None
    self.total_num_searches = 0
    self.total_search_time = 0
    # Async MCTS parameters
    self.virtual_loss = virtual_loss
    self.batch_size = batch_size
    self.timeout = timeout
    self.total_timeouts = 0
    self.total_eval_errors = 0
    # Child selection at root parameters
    self._secure_c = secure_c
    self._alternative_criteria = False
    self._max_additional_simulation_rounds = max_additional_simulation_rounds
    self._simulations_multiplier = simulations_multiplier

  def restart_at(self, state):
    pass

  def get_root(self):
    return self._root

  def _get_selection_function_arguments(self, root):
    if self._best_child_fn is secure_child:
      arguments = (root, self._secure_c)
    elif self._best_child_fn is max_robust_child:
      arguments = (root, self._alternative_criteria)
    elif self._best_child_fn is max_robust_secure_child:
      arguments = (root, self._secure_c, self._alternative_criteria)
    else:
      arguments = (root,)

    return arguments

  def step_with_policy(self, state):
    """Returns bot's policy and action at given state."""
    t1 = time.time()
    simulation_round = 0
    best_action = None
    best_child = None
    root = None
    while (
        best_action is None
        and simulation_round <= self._max_additional_simulation_rounds
    ):
      simulation_round += 1
      if simulation_round == self._max_additional_simulation_rounds:
        self._alternative_criteria = True
      root = self.mcts_search(state)
      assert root is not None, "Root is None"
      self._root = root
      arguments = self._get_selection_function_arguments(root)
      best_action, best_child = self._best_child_fn(*arguments)
      # Determine the number of additional simulations.
      if best_action is None:
        self.max_simulations = (
            int(self._simulations_multiplier * self._max_simulations)
        )
    assert best_action is not None, "Best action is None"
    assert best_child is not None, "Best child is None"
    assert self._root is not None, "Root is None"
    seconds = time.time() - t1
    self.total_search_time += seconds
    if self.verbose:
      print(
          "Finished {} sims in {:.3f} secs, {:.1f} sims/s".format(
              root.explore_count, seconds, root.explore_count / seconds
          )
      )
      print("Root:")
      print(root.to_str(state))
      print("Children:")
      print(root.children_str(state))
      if best_child.children:
        chosen_state = state.clone()
        chosen_state.apply_action(best_action)
        print("Children of chosen:")
        print(best_child.children_str(chosen_state))
    policy = [
        (action, (1.0 if action == best_action else 0.0))
        for action in state.legal_actions(state.current_player())
    ]
    # Rest max simulations to original value.
    self.max_simulations = self._max_simulations
    return policy, best_action

  def step(self, state):
    return self.step_with_policy(state)[1]

  def _add_virtual_losses(self, node):
    # Add virtual losses. This is applied to the nodes touched during the tree
    # policy action selection of the simulation (downward pass). This
    # discourages multiple threads from exploring the same node.
    node.total_reward += (self.virtual_loss * self.min_utility)
    node.explore_count += self.virtual_loss

  def _remove_virtual_losses(self, node):
    # Remove virtual losses. This is applied to the nodes touched during the
    # backpropagation phase (upward pass) to ensure that the fake losses are
    # removed once this path down the tree is done being simulated.
    node.total_reward -= (self.virtual_loss * self.min_utility)
    node.explore_count -= self.virtual_loss

  def _choose_next_node(self, visit_path, working_state, current_node):
    if working_state.is_chance_node():
      # For chance nodes, rollout according to chance node's probability
      # distribution
      outcomes = working_state.chance_outcomes()
      action_list, prob_list = zip(*outcomes)
      action = self._random_state.choice(action_list, p=prob_list)
      chosen_child = next(
          c for c in current_node.children if c.action == action
      )
    else:
      # Otherwise choose node with largest UCT value
      chosen_child = max(
          current_node.children,
          key=lambda c: self._child_selection_fn(  # pylint: disable=g-long-lambda
              c, current_node.explore_count, self.uct_c
          ),
      )
    working_state.apply_action(chosen_child.action)
    current_node = chosen_child
    self._add_virtual_losses(current_node)
    visit_path.append(current_node)
    return current_node

  def _apply_tree_policy(self, root, state):
    """Applies the UCT policy to play the game until reaching a leaf node.

    A leaf node is defined as a node that is terminal or has not been evaluated
    yet. If it reaches a node that has been evaluated before but hasn't been
    expanded, then expand its children and continue.

    Args:
      root: The root node in the search tree.
      state: The state of the game at the root node.

    Returns:
      visit_path: A list of nodes descending from the root node to a leaf node.
      working_state: The state of the game at the leaf node.
    """
    visit_path = [root]
    working_state = state.clone()
    current_node = root
    self._add_virtual_losses(root)
    unexplored_explore_count = self.virtual_loss
    while (
        not working_state.is_terminal()
        and current_node.explore_count > unexplored_explore_count
    ) or (working_state.is_chance_node() and self.dont_return_chance_node):
      if not current_node.children:
        return visit_path, working_state, current_node
      current_node = self._choose_next_node(
          visit_path, working_state, current_node
      )
    return visit_path, working_state, current_node

  def backpropagate(self, visit_path, returns):
    while visit_path:
      # For chance nodes, walk up the tree to find the decision-maker.
      decision_node_idx = -1
      while visit_path[decision_node_idx].player == pyspiel.PlayerId.CHANCE:
        decision_node_idx -= 1
      # Chance node targets are for the respective decision-maker.
      target_return = returns[visit_path[decision_node_idx].player]
      node = visit_path.pop()
      node.total_reward += target_return
      node.explore_count += 1
      self._remove_virtual_losses(node)
      assert node.explore_count >= 1

  def backpropagate_timeout(self, visit_path):
    while visit_path:
      # For chance nodes, walk up the tree to find the decision-maker.
      node = visit_path.pop()
      self._remove_virtual_losses(node)

  def expand(self, root, working_state, current_node, prior, value):
    # For a new node, initialize its state, then choose a child as normal.
    # prior, value = self.evaluator.prior_and_value(working_state)
    if current_node is root and self._dirichlet_noise:
      epsilon, alpha = self._dirichlet_noise
      noise = self._random_state.dirichlet([alpha] * len(prior))
      prior = [
          (a, (1 - epsilon) * p + epsilon * n)
          for (a, p), n in zip(prior, noise)
      ]
    # Reduce bias from move generation order.
    self._random_state.shuffle(prior)
    player = working_state.current_player()
    current_node.children = [
        SearchNode(action, player, prob) for action, prob in prior
    ]
    current_node.expanded = True

  def evaluate(
      self, working_state
  ) -> tuple[list[tuple[int, float]], np.ndarray]:
    if working_state.is_terminal():
      prior = []
      values = working_state.returns()
    else:
      prior, values = self.evaluator.prior_and_value(working_state)
    return prior, values

  def handle_leaf(self, prior, value, arguments, timeout=False):
    visit_path, working_state, node, root = arguments
    if timeout:
      self.backpropagate_timeout(visit_path)
      return
    assert node is not None
    if not node.expanded:
      self.expand(root, working_state, node, prior, value)
    if working_state.is_terminal():
      visit_path[-1].outcome = working_state.returns()
    self.backpropagate(visit_path, value)

  def async_mcts_search(self, state):
    root = SearchNode(None, state.current_player(), 1)
    # do one call up front, to ensure we have some children for the threads
    # to spread over.
    self._add_virtual_losses(root)
    working_state = state.clone()
    prior, value = self.evaluate(working_state)
    self.handle_leaf(
        prior, value, ([root], working_state, root, root), timeout=False
    )
    total_simulations = 1
    search_timeouts = 0
    with concurrent.futures.ThreadPoolExecutor(
        max_workers=self.batch_size
    ) as executor:
      while total_simulations < self.max_simulations:
        batch_timeouts = 0
        remaining_simulations = self.max_simulations - total_simulations
        num_to_queue = min(self.batch_size, remaining_simulations)
        futures = []
        arguments = []
        for _ in range(num_to_queue):
          visit_path, working_state, node = self._apply_tree_policy(root, state)
          arguments.append((visit_path, working_state, node, root))
          future = executor.submit(self.evaluate, working_state)
          futures.append(future)
        concurrent.futures.wait(
            futures,
            timeout=self.timeout,
            return_when=concurrent.futures.ALL_COMPLETED,
        )
        for i, future in enumerate(futures):
          if future.done():
            prior, value = future.result()
            self.handle_leaf(prior, value, arguments[i], timeout=False)
          else:
            batch_timeouts += 1
            search_timeouts += 1
            self.handle_leaf(None, None, arguments[i], timeout=True)
        total_simulations += num_to_queue
    if self.verbose:
      logging.info("Timeouts for this search: %d", search_timeouts)
    self.total_timeouts += search_timeouts
    self.total_num_searches += 1
    if self.verbose:
      logging.info(
          "Average timeouts per search: %g\n" +
          "Average eval errors per search: %g",
          self.total_timeouts / self.total_num_searches,
          self.total_eval_errors / self.total_num_searches)
    return root

  def mcts_search(self, state):
    """A vanilla Monte-Carlo Tree Search algorithm.

    This algorithm searches the game tree from the given state.
    At the leaf, the evaluator is called if the game state is not terminal.
    A total of max_simulations states are explored.

    At every node, the algorithm chooses the action with the highest PUCT value,
    defined as: `Q/N + c * prior * sqrt(parent_N) / N`, where Q is the total
    reward after the action, and N is the number of times the action was
    explored in this position. The input parameter c controls the balance
    between exploration and exploitation; higher values of c encourage
    exploration of under-explored nodes. Unseen actions are always explored
    first.

    At the end of the search, the chosen action is the action that has been
    explored most often. This is the action that is returned.

    This implementation supports sequential n-player games, with or without
    chance nodes. All players maximize their own reward and ignore the other
    players' rewards. This corresponds to max^n for n-player games. It is the
    norm for zero-sum games, but doesn't have any special handling for
    non-zero-sum games. It doesn't have any special handling for imperfect
    information games.

    The implementation also supports backing up solved states, i.e. MCTS-Solver.
    The implementation is general in that it is based on a max^n backup (each
    player greedily chooses their maximum among proven children values, or there
    exists one child whose proven value is game.max_utility()), so it will work
    for multiplayer, general-sum, and arbitrary payoff games (not just win/loss/
    draw games). Also chance nodes are considered proven only if all children
    have the same value.

    Some references:
    - Sturtevant, An Analysis of UCT in Multi-Player Games,  2008,
      https://web.cs.du.edu/~sturtevant/papers/multi-player_UCT.pdf
    - Nijssen, Monte-Carlo Tree Search for Multi-Player Games, 2013,
      https://project.dke.maastrichtuniversity.nl/games/files/phd/Nijssen_thesis.pdf
    - Silver, AlphaGo Zero: Starting from scratch, 2017
      https://deepmind.com/blog/article/alphago-zero-starting-scratch
    - Winands, Bjornsson, and Saito, "Monte-Carlo Tree Search Solver", 2008.
      https://dke.maastrichtuniversity.nl/m.winands/documents/uctloa.pdf

    Arguments:
      state: pyspiel.State object, state to search from

    Returns:
      The most visited move from the root node.
    """
    return self.async_mcts_search(state)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example to traverse an entire game tree."""

from absl import app
from absl import flags

import numpy as np
from open_spiel.python import games  # pylint: disable=unused-import
from open_spiel.python.bots import human
from open_spiel.python.bots import uniform_random
import pyspiel

_GAME_STRING = flags.DEFINE_string(
    "game_string", "tic_tac_toe", "Name of the game"
)
_PLAYER0_TYPE = flags.DEFINE_string(
    "player0_type", "human", "Player 0 type (human or uniform)"
)
_PLAYER1_TYPE = flags.DEFINE_string(
    "player1_type", "uniform", "Player 1 type (human or uniform)"
)


def load_bot(bot_type: str, pid: int) -> pyspiel.Bot:
  if bot_type == "human":
    return human.HumanBot()
  elif bot_type == "uniform":
    return uniform_random.UniformRandomBot(pid, np.random)


def play_game(state: pyspiel.State,
              bots: list[pyspiel.Bot]):
  """Play the game via console."""

  while not state.is_terminal():
    print(f"State: \n{state}\n")
    if state.is_chance_node():
      outcomes = state.chance_outcomes()
      action_list, prob_list = zip(*outcomes)
      outcome = np.random.choice(action_list, p=prob_list)
      print(f"Chance chose: {outcome} ({state.action_to_string(outcome)})")
      state.apply_action(outcome)
    else:
      player = state.current_player()
      action = bots[player].step(state)
      print(f"Chose action: {action} ({state.action_to_string(action)})")
      state.apply_action(action)

  print("\n-=- Game over -=-\n")
  print(f"Terminal state:\n{state}")
  print(f"Returns: {state.returns()}")
  return


def main(_):
  game = pyspiel.load_game(_GAME_STRING.value)
  state = game.new_initial_state()
  bots = []
  bots.append(load_bot(_PLAYER0_TYPE.value, 0))
  bots.append(load_bot(_PLAYER1_TYPE.value, 1))
  play_game(state, bots)


if __name__ == "__main__":
  app.run(main)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.voting.ranked_pairs."""

from absl.testing import absltest
import numpy as np
from open_spiel.python.voting import base
from open_spiel.python.voting import ranked_pairs


class RankedPairsTest(absltest.TestCase):

  def test_ranked_pairs_wikipedia_example1(self):
    alternatives = ["w", "x", "y", "z"]
    votes = [
        base.WeightedVote(7, ["w", "x", "z", "y"]),
        base.WeightedVote(2, ["w", "y", "x", "z"]),
        base.WeightedVote(4, ["x", "y", "z", "w"]),
        base.WeightedVote(5, ["x", "z", "w", "y"]),
        base.WeightedVote(1, ["y", "w", "x", "z"]),
        base.WeightedVote(8, ["y", "z", "w", "x"]),
    ]
    profile = base.PreferenceProfile(votes=votes, alternatives=alternatives)
    method = ranked_pairs.RankedPairsVoting()
    outcome = method.run_election(profile)
    with self.subTest("Ranking and scores"):
      self.assertListEqual(outcome.ranking, ["w", "x", "y", "z"])
      self.assertListEqual(outcome.scores, [29, 19, 3, 0])
    with self.subTest("Check the graph"):
      expected_graph = np.array(
          [[0, 1, 1, 0], [0, 0, 1, 1], [0, 0, 0, 1], [0, 0, 0, 0]]
      )
      self.assertTrue(np.array_equal(outcome.graph, expected_graph))

  def test_ranked_pairs_wikipedia_example2(self):
    alternatives = ["Memphis", "Nashville", "Chattanooga", "Knoxville"]
    votes = [
        base.WeightedVote(
            42, ["Memphis", "Nashville", "Chattanooga", "Knoxville"]
        ),
        base.WeightedVote(
            26, ["Nashville", "Chattanooga", "Knoxville", "Memphis"]
        ),
        base.WeightedVote(
            15, ["Chattanooga", "Knoxville", "Nashville", "Memphis"]
        ),
        base.WeightedVote(
            17, ["Knoxville", "Chattanooga", "Nashville", "Memphis"]
        ),
    ]
    profile = base.PreferenceProfile(votes=votes, alternatives=alternatives)
    method = ranked_pairs.RankedPairsVoting()
    outcome = method.run_election(profile)
    with self.subTest("Ranking and scores"):
      self.assertListEqual(
          outcome.ranking, ["Nashville", "Chattanooga", "Knoxville", "Memphis"]
      )
      self.assertListEqual(outcome.scores, [186, 98, 16, 0])
    with self.subTest("Check the graph"):
      expected_graph = np.array(
          [[0, 0, 0, 0], [1, 0, 1, 1], [1, 0, 0, 1], [1, 0, 0, 0]]
      )
      self.assertTrue(np.array_equal(outcome.graph, expected_graph))

  def test_meeple_pentathlon(self):
    alternatives = ["A", "B", "C"]
    votes = [
        base.WeightedVote(1, ["A", "B", "C"]),
        base.WeightedVote(1, ["A", "C", "B"]),
        base.WeightedVote(2, ["C", "A", "B"]),
        base.WeightedVote(1, ["B", "C", "A"]),
    ]
    profile = base.PreferenceProfile(votes=votes, alternatives=alternatives)
    method = ranked_pairs.RankedPairsVoting()
    outcome = method.run_election(profile)
    with self.subTest("Ranking and scores"):
      self.assertListEqual(outcome.ranking, ["C", "A", "B"])
      self.assertListEqual(outcome.scores, [5, 3, 0])
    with self.subTest("Check the graph"):
      # A -> B, C -> A, C -> B
      expected_graph = np.array([[0, 1, 0], [0, 0, 0], [1, 1, 0]])
      self.assertTrue(np.array_equal(outcome.graph, expected_graph))

  def test_ranked_pairs_simple_cycle(self):
    alternatives = ["A", "B"]
    votes = [
        base.WeightedVote(1, ["A", "B"]),
        base.WeightedVote(1, ["B", "A"]),
    ]
    profile = base.PreferenceProfile(votes=votes, alternatives=alternatives)
    method = ranked_pairs.RankedPairsVoting()
    outcome = method.run_election(profile)
    with self.subTest("Check the graph is empty"):
      expected_graph = np.array(
          [[0, 0], [0, 0]]
      )
      self.assertTrue(np.array_equal(outcome.graph, expected_graph))
    with self.subTest("Rankings and scores"):
      self.assertTrue(outcome.ranking == ["A", "B"] or
                      outcome.ranking == ["B", "A"])
      self.assertListEqual(outcome.scores, [0, 0])

if __name__ == "__main__":
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Regret matching."""
from absl import flags
import jax
import jax.numpy as jnp
import numpy as np

FLAGS = flags.FLAGS


class RegretMatchingAgent:
  """Regret matching agent."""

  def __init__(self, num_actions, data_loader):
    self.num_actions = num_actions
    # self.regret_sum = jax.numpy.array(np.zeros(self.num_actions))
    self.regret_sum = jax.numpy.array(
        np.zeros(shape=[FLAGS.batch_size, 1, self.num_actions]))
    self.data_loader = data_loader

  def train(self):
    pass

  def initial_policy(self):
    self.last_policy = self.regret_matching_policy(self.regret_sum)
    return self.last_policy

  def next_policy(self, last_values):
    value = jnp.matmul(self.last_policy, last_values)
    last_values = jnp.transpose(last_values, [0, 2, 1])
    current_regrets = last_values - value
    self.regret_sum += current_regrets
    self.last_policy = self.regret_matching_policy(self.regret_sum)
    return self.last_policy

  def regret_matching_policy(self, regret_sum):
    """Regret matching policy."""

    strategy = np.copy(regret_sum)
    strategy[strategy < 0] = 0
    strategy_sum = np.sum(strategy, axis=-1)
    for i in range(FLAGS.batch_size):
      if strategy_sum[i] > 0:
        strategy[i] /= strategy_sum[i]
      else:
        strategy[i] = np.repeat(1 / self.num_actions, self.num_actions)
    return strategy



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.voting.kemeny_young."""

from absl.testing import absltest

from open_spiel.python.voting import base
from open_spiel.python.voting import kemeny_young


class KemenyYoungTest(absltest.TestCase):

  def test_ranked_pairs_wikipedia_example(self):
    alternatives = ["Memphis", "Nashville", "Chattanooga", "Knoxville"]
    votes = [
        base.WeightedVote(42,
                          ["Memphis", "Nashville", "Chattanooga", "Knoxville"]),
        base.WeightedVote(26,
                          ["Nashville", "Chattanooga", "Knoxville", "Memphis"]),
        base.WeightedVote(15,
                          ["Chattanooga", "Knoxville", "Nashville", "Memphis"]),
        base.WeightedVote(17,
                          ["Knoxville", "Chattanooga", "Nashville", "Memphis"]),
    ]
    profile = base.PreferenceProfile(votes=votes, alternatives=alternatives)
    method = kemeny_young.KemenyYoungVoting()
    outcome = method.run_election(profile)
    self.assertListEqual(outcome.ranking,
                         ["Nashville", "Chattanooga", "Knoxville", "Memphis"])
    self.assertListEqual(outcome.scores, [194, 141, 58, 0])

  def test_meeple_pentathlon(self):
    alternatives = ["A", "B", "C"]
    votes = [
        base.WeightedVote(1, ["A", "B", "C"]),
        base.WeightedVote(1, ["A", "C", "B"]),
        base.WeightedVote(2, ["C", "A", "B"]),
        base.WeightedVote(1, ["B", "C", "A"]),
    ]
    profile = base.PreferenceProfile(votes=votes, alternatives=alternatives)
    method = kemeny_young.KemenyYoungVoting()
    outcome = method.run_election(profile)
    self.assertListEqual(outcome.ranking, ["C", "A", "B"])
    self.assertListEqual(outcome.scores, [6, 4, 0])


if __name__ == "__main__":
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for constructing strings in color."""

RESET = '\033[0m'         # Reset
BLACK = '\033[30m'        # Black
RED = '\033[31m'          # Red -- Terminating Game
GREEN = '\033[32m'        # Green -- Computing Payoffs
YELLOW = '\033[33m'       # Yellow -- Generated Game Def
BLUE = '\033[34m'         # Blue
PURPLE = '\033[35m'       # Purple -- Information States
CYAN = '\033[36m'         # Cyan -- Generating Lists
WHITE = '\033[37m'        # White
BLACK2 = '\033[39m'       # Black?


class ColorText:
  """Color text class."""

  def __init__(self, reset_color=RESET):
    self.reset_color = reset_color
    self.current_color = reset_color

  def set_color(self, color: str):
    self.current_color = color

  def set_reset_color(self, color: str):
    self.reset_color = color

  def reset(self):
    self.current_color = self.reset_color

  def color(self, log_str: str, color: str = ''):
    c = color if color else self.current_color
    log_str = c + log_str + self.reset_color
    return log_str



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Dominoes (4 players) implemented in Python.

https://en.wikipedia.org/wiki/Dominoes#Latin_American_Version
"""

import collections
import copy
import itertools

import numpy as np

import pyspiel

_NUM_PLAYERS = 4
_PIPS = [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0]
_DECK = list(itertools.combinations_with_replacement(_PIPS, 2))
_EDGES = [None, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0]


class Action:
  """Represent player possible action."""

  def __init__(self, player, tile, edge):
    self.player = player
    self.tile = tile
    self.edge = edge

  def __str__(self):
    return f"p{self.player} tile:{self.tile} pip:{self.edge}"

  def __repr__(self):
    return self.__str__()


def create_possible_actions():
  actions = []
  for player in range(_NUM_PLAYERS):
    for tile in _DECK:
      for edge in _EDGES:
        if edge in tile or edge is None:
          actions.append(Action(player, tile, edge))
  return actions


_ACTIONS = create_possible_actions()
_ACTIONS_STR = [str(action) for action in _ACTIONS]

_HAND_SIZE = 7

_MAX_GAME_LENGTH = 28

_GAME_TYPE = pyspiel.GameType(
    short_name="python_team_dominoes",
    long_name="Python Team Dominoes (4 players)",
    dynamics=pyspiel.GameType.Dynamics.SEQUENTIAL,
    chance_mode=pyspiel.GameType.ChanceMode.EXPLICIT_STOCHASTIC,
    information=pyspiel.GameType.Information.IMPERFECT_INFORMATION,
    utility=pyspiel.GameType.Utility.ZERO_SUM,
    reward_model=pyspiel.GameType.RewardModel.TERMINAL,
    max_num_players=_NUM_PLAYERS,
    min_num_players=_NUM_PLAYERS,
    provides_information_state_string=True,
    provides_information_state_tensor=True,
    provides_observation_string=True,
    provides_observation_tensor=True,
    provides_factored_observation_string=True,
)
_GAME_INFO = pyspiel.GameInfo(
    num_distinct_actions=len(_ACTIONS),
    max_chance_outcomes=len(_DECK),
    min_utility=-100,
    max_utility=100,
    num_players=_NUM_PLAYERS,
    # deal: 28 chance nodes + play: 28 player nodes
    max_game_length=_MAX_GAME_LENGTH,
    utility_sum=0.0,
)


class DominoesGame(pyspiel.Game):
  """A Python version of Block Dominoes."""

  def __init__(self, params=None):
    super().__init__(_GAME_TYPE, _GAME_INFO, params or dict())

  def new_initial_state(self):
    """Returns a state corresponding to the start of a game."""
    return DominoesState(self)

  def make_py_observer(self, iig_obs_type=None, params=None):
    """Returns an object used for observing game state."""
    return DominoesObserver(
        iig_obs_type or pyspiel.IIGObservationType(perfect_recall=False), params
    )


class DominoesState(pyspiel.State):
  """A python version of the Block Dominoes state."""

  def __init__(self, game):
    """Constructor; should only be called by Game.new_initial_state."""
    super().__init__(game)
    self.actions_history = []
    self.open_edges = []
    self.hands = [[] for _ in range(_NUM_PLAYERS)]
    self.deck = copy.deepcopy(_DECK)
    self._game_over = False
    self._next_player = pyspiel.PlayerId.CHANCE
    self._current_deal_player = 0  # NEW ATTRIBUTE

  # OpenSpiel (PySpiel) API functions are below. This is the standard set that
  # should be implemented by every sequential-move game with chance.

  def current_player(self):
    """Returns id of the next player to move, or TERMINAL if game is over."""
    if self._game_over:
      return pyspiel.PlayerId.TERMINAL
    if self.deck:  # deal phase
      return pyspiel.PlayerId.CHANCE
    return self._next_player

  def _legal_actions(self, player):
    """Returns a list of legal actions, sorted in ascending order."""
    assert player >= 0
    assert player == self._next_player
    return self.get_legal_actions(player)

  def get_legal_actions(self, player):
    """Returns a list of legal actions."""
    assert player >= 0

    actions = []
    hand = self.hands[player]

    # first move, no open edges
    if not self.open_edges:
      for tile in hand:
        actions.append(Action(player, tile, None))
    else:
      for tile in hand:
        if tile[0] in self.open_edges:
          actions.append(Action(player, tile, tile[0]))
        if tile[0] != tile[1] and tile[1] in self.open_edges:
          actions.append(Action(player, tile, tile[1]))

    actions_idx = [_ACTIONS_STR.index(str(action)) for action in actions]
    actions_idx.sort()
    return actions_idx

  def chance_outcomes(self):
    """Returns the possible chance outcomes and their probabilities."""
    assert self.is_chance_node()
    p = 1.0 / len(self.deck)
    return [(_DECK.index(i), p) for i in self.deck]

  def _apply_action(self, action):
    """Applies the specified action to the state."""
    if self.is_chance_node():
      # Deal tiles to players in order (0, 1, 2, 3)
      hand_to_add_tile = self.hands[self._current_deal_player]
      tile = _DECK[action]
      self.deck.remove(tile)
      hand_to_add_tile.append(tile)
      self._current_deal_player = (self._current_deal_player + 1) % 4

      # Check if all hands are of _HAND_SIZE
      if not all(len(hand) == _HAND_SIZE for hand in self.hands):
        return  # more tiles to deal

      for hand in self.hands:
        hand.sort()

      self._next_player = 0
    else:
      action = _ACTIONS[action]
      self.actions_history.append(action)
      my_idx = self.current_player()
      my_hand = self.hands[my_idx]
      my_hand.remove(action.tile)
      self.update_open_edges(action)

      if not my_hand:
        self._game_over = True  # player played his last tile
        return

      for i in range(1, 5):
        next_idx = (my_idx + i) % 4
        next_legal_actions = self.get_legal_actions(next_idx)

        if next_legal_actions:
          self._next_player = next_idx
          return

        # Check if a team has played all their tiles.
        if not (self.hands[0] or self.hands[2]) or not (
            self.hands[1] or self.hands[3]
        ):
          self._game_over = True
          return

      # all players are blocked. Game is stuck.
      self._game_over = True

  def update_open_edges(self, action):
    if not self.open_edges:
      self.open_edges = list(action.tile)
    else:
      self.open_edges.remove(action.edge)
      new_edge = (
          action.tile[0] if action.tile[0] != action.edge else action.tile[1]
      )
      self.open_edges.append(new_edge)

    self.open_edges.sort()

  def _action_to_string(self, player, action):
    """Action -> string."""
    if player == pyspiel.PlayerId.CHANCE:
      return f"Deal {_DECK[action]}"
    return _ACTIONS_STR[action]

  def is_terminal(self):
    """Returns True if the game is over."""
    return self._game_over

  def returns(self):
    """Total reward for each player over the course of the game so far."""
    if not self.is_terminal():
      return [0 for _ in range(_NUM_PLAYERS)]

    sum_of_pips0 = sum(t[0] + t[1] for t in (self.hands[0] + self.hands[2]))
    sum_of_pips1 = sum(t[0] + t[1] for t in (self.hands[1] + self.hands[3]))

    if sum_of_pips1 == sum_of_pips0:
      return [0 for _ in range(_NUM_PLAYERS)]

    if sum_of_pips1 > sum_of_pips0:
      return [sum_of_pips1, -sum_of_pips1, sum_of_pips1, -sum_of_pips1]
    return [-sum_of_pips0, sum_of_pips0, -sum_of_pips0, sum_of_pips0]

  def __str__(self):
    """String for debug purposes. No particular semantics are required."""
    hand0 = [str(c) for c in self.hands[0]]
    hand1 = [str(c) for c in self.hands[1]]
    hand2 = [str(c) for c in self.hands[2]]
    hand3 = [str(c) for c in self.hands[3]]
    board = self.draw_board()
    return (
        f"hand0:{hand0}\n"
        f"hand1:{hand1}\n"
        f"hand2:{hand2}\n"
        f"hand3:{hand3}\n\n"
        f"board: {board}"
    )

  def draw_board(self):
    """Draw the board' in a human readable format."""
    board = collections.deque()
    current_open_edges = None
    for action in self.actions_history:
      # check if action is played on an empty board
      if action.edge is None:
        board.append(action.tile)
        # pylint: disable=unused-variable
        current_open_edges = list(action.tile)
      # check if action edge matches last played edge in the left or right
      elif action.edge == current_open_edges[0]:
        # invert the tile if the edge is on the right:
        tile = (
            (action.tile[1], action.tile[0])
            if action.tile[0] == current_open_edges[0]
            else action.tile
        )
        board.appendleft(tile)

      elif action.edge == current_open_edges[1]:
        # invert the tile if the edge is on the left:
        tile = (
            (action.tile[1], action.tile[0])
            if action.tile[1] == current_open_edges[1]
            else action.tile
        )
        board.append(tile)

      current_open_edges = board[0][0], board[-1][1]

    # TODO(someone): move this to a test
    assert len(board) == len(self.actions_history)
    return list(board)


class DominoesObserver:
  """Observer, conforming to the PyObserver interface (see observation.py)."""

  def __init__(self, iig_obs_type, params):
    """Initializes an empty observation tensor."""
    if params:
      raise ValueError(f"Observation parameters not supported; passed {params}")

    # Determine which observation pieces we want to include.
    pieces = [("player", 4, (4,))]

    if iig_obs_type.private_info == pyspiel.PrivateInfoType.SINGLE_PLAYER:
      # each tile is represented using 3 integers:
      # 2 for the pips, and 1 to distinguish between (0,0) to empty slot for
      # a tile.
      pieces.append(("hand", 21, (7, 3)))  # 7 tiles per hand
    if iig_obs_type.public_info:
      if iig_obs_type.perfect_recall:
        # list of all played actions, each action is represented using 5
        # integers:
        # 2 for the played tile (0-6),
        # 1 for the covered edge (0-6),
        # 1 for which player (0,1,3,4),
        # 1 to distinguish between actual move and empty slot for a move (0/1).
        # the None (play on an empty board) edge represented using 0.
        pieces.append(("actions_history", 125, (25, 5)))
      else:
        # last action, represented in the same way as in "actions_history"
        # but without the last integer.
        pieces.append(("last_action", 4, (4,)))
        pieces.append(("hand_sizes", 4, (4,)))

    # Build the single flat tensor.
    total_size = sum(size for name, size, shape in pieces)
    self.tensor = np.zeros(total_size, np.float32)

    # Build the named & reshaped views of the bits of the flat tensor.
    self.dict = {}
    index = 0
    for name, size, shape in pieces:
      self.dict[name] = self.tensor[index : index + size].reshape(shape)
      index += size

  def copy_indices(self, dest, source, index_list):
    for idx in index_list:
      dest[idx] = source[idx]

  def set_from(self, state, player):
    """Updates `tensor` and `dict` to reflect `state` from PoV of `player`."""

    self.tensor.fill(0)

    if "player" in self.dict:
      self.dict["player"][player] = 1
      self.dict["player"][1 - player] = 0

    if "hand_sizes" in self.dict:
      my_hand_size = len(state.hands[player])
      opp_hand_size = len(state.hands[1 - player])
      self.dict["hand_sizes"][0] = my_hand_size
      self.dict["hand_sizes"][1] = opp_hand_size

    if "edges" in self.dict:
      if state.open_edges:
        self.copy_indices(self.dict["edges"], state.open_edges, [0, 1])
      else:
        self.dict["edges"][0] = 0.0
        self.dict["edges"][1] = 0.0

    if "hand" in self.dict:
      for i, tile in enumerate(state.hands[player]):
        self.copy_indices(self.dict["hand"][i], tile, [0, 1])
        self.dict["hand"][i][2] = 1.0

    if "actions_history" in self.dict:
      for i, action in enumerate(state.actions_history):
        self.copy_indices(self.dict["actions_history"][i], action.tile, [0, 1])
        self.dict["actions_history"][i][2] = (
            action.edge if action.edge is not None else 0.0
        )
        self.dict["actions_history"][i][3] = action.player
        self.dict["actions_history"][i][4] = 1.0

    if "last_action" in self.dict:
      if state.actions_history:
        action = state.actions_history[-1]
        self.copy_indices(self.dict["last_action"], action.tile, [0, 1])
        self.dict["last_action"][2] = (
            action.edge if action.edge is not None else 0.0
        )
        self.dict["last_action"][3] = action.player

  def string_from(self, state, player):
    """Observation of `state` from the PoV of `player`, as a string."""
    pieces = []
    if "player" in self.dict:
      pieces.append(f"p{player}")
    if "hand" in self.dict:
      pieces.append(f"hand:{state.hands[player]}")
    if "actions_history" in self.dict:
      pieces.append(f"history:{str(state.actions_history)}")
    if "last_action" in self.dict and state.actions_history:
      pieces.append(f"last_action:{str(state.actions_history[-1])}")
    return " ".join(str(p) for p in pieces)


# Register the game with the OpenSpiel library

pyspiel.register_game(_GAME_TYPE, DominoesGame)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Mean field proximal policy optimaztion."""

# pylint: disable=consider-using-from-import
# pylint: disable=g-importing-member

from absl.testing import absltest
from absl.testing import parameterized
from open_spiel.python.mfg.algorithms.pytorch.mfg_proximal_policy_optimization import Agent as mfg_ppo_agent
from open_spiel.python.mfg.algorithms.pytorch.mfg_proximal_policy_optimization import calculate_advantage
from open_spiel.python.mfg.algorithms.pytorch.mfg_proximal_policy_optimization import calculate_explotability
from open_spiel.python.mfg.algorithms.pytorch.mfg_proximal_policy_optimization import learn
from open_spiel.python.mfg.algorithms.pytorch.mfg_proximal_policy_optimization import Policy as mfg_ppo_policy
from open_spiel.python.mfg.algorithms.pytorch.mfg_proximal_policy_optimization import rollout
import torch
import torch.optim as optim

from open_spiel.python import policy as policy_std
from open_spiel.python import rl_environment
from open_spiel.python.mfg.algorithms import distribution
from open_spiel.python.mfg.games import factory


class PolicyTest(parameterized.TestCase):
  """Test the policy."""

  @parameterized.named_parameters(
      ("python", "mfg_crowd_modelling_2d", "crowd_modelling_2d_four_rooms")
  )
  def test_train(self, name, setting):
    """Checks that the training works."""
    device = torch.device("cpu")
    args = {
        "num_episodes": 5,
        "gamma": 0.9,
    }
    game = factory.create_game_with_setting(name, setting)
    uniform_policy = policy_std.UniformRandomPolicy(game)
    mfg_dist = distribution.DistributionPolicy(game, uniform_policy)
    env = rl_environment.Environment(
        game, mfg_distribution=mfg_dist, mfg_population=0
    )

    # Set the environment seed for reproduciblility
    env.seed(0)

    # Creat the agent and population policies
    info_state_size = env.observation_spec()["info_state"][0]
    num_actions = env.action_spec()["num_actions"]

    agent = mfg_ppo_agent(info_state_size, num_actions).to(device)
    ppo_policy = mfg_ppo_policy(game, agent, None, device)
    pop_agent = mfg_ppo_agent(info_state_size, num_actions).to(device)

    optimizer_actor = optim.Adam(agent.actor.parameters(), lr=1e-3, eps=1e-5)
    optimizer_critic = optim.Adam(agent.critic.parameters(), lr=1e-3, eps=1e-5)

    # calculate the exploitability
    m = calculate_explotability(game, mfg_dist, ppo_policy)
    init_nashc = m["nash_conv_ppo"]

    steps = args["num_episodes"] * env.max_game_length

    for _ in range(3):
      # collect rollout data
      history = rollout(
          env, pop_agent, agent, args["num_episodes"], steps, device
      )
      # Calculate the advantage function
      adv, returns = calculate_advantage(
          args["gamma"],
          True,
          history["rewards"],
          history["values"],
          history["dones"],
          device,
      )
      history["advantages"] = adv
      history["returns"] = returns
      # Update the learned policy and report loss for debugging
      learn(history, optimizer_actor, optimizer_critic, agent)

    # Update the iteration policy with the new policy
    pop_agent.load_state_dict(agent.state_dict())

    # Update the distribution
    distrib = distribution.DistributionPolicy(game, ppo_policy)

    # calculate the exploitability
    m = calculate_explotability(game, distrib, ppo_policy)
    nashc = m["nash_conv_ppo"]

    # update the environment distribution
    env.update_mfg_distribution(distrib)

    # Test convergence
    self.assertLessEqual(nashc, 2 * init_nashc)


if __name__ == "__main__":
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Find ratings by minimizing a differentiable Kendall-tau distance.

This is an idea for an optimization algorithm motivated by work on
voting-as-evaluation (VasE); see go/gdm-vase.  Please contact lanctot@ if you
have any comments or questions.

This method was first inspired by Condorcet's vision that there is an underlying
true ranking and that voting rules are simply noisy estimates of the ground
truth rank. As shown in Section 8.3 of the Handbook of Computational Social
Choice (https://www.cse.unsw.edu.au/~haziz/comsoc.pdf) the maximum likelihood
estimate leads to a ranking that minimized the Kendall-tau distance to the
votes, which is precisely what the Kemeny-Young voting method does. However,
its complexity is O(m!) where m is the number of alternatives. Also, it is not
clear how the social choice methods handle sparsity in the ranking data.

This method, Soft Condorcet Optimization (SCO), assigns a numerical rating to
each alternative: r_i, and defines the loss to be:

    sum_{v in Votes} sum_{alternatives a, b in v s.t. a > b} D(r_a, r_b)

If D is defined to be:
        0  if r_a - r_b > 0    (r_a > r_b -> correct ordering)
  or    1  otherwise           (r_b >= r_a -> incorrect ordering)

then the loss above is the sum of Kendall-tau distances and the minimum
corresponds to the solution Kemeny-Young would find. But, it's not
differentiable. In SCO, we replace D by a sigmoid (smooth step function).

   D(r_a, r_b) = sigmoid((r_b - r_a) / tau)
               = sigmoid(Delta_{ab})

where sigmoid(x) = 1.0 / (1.0 + exp(-x)). The partial derivatives of D(r_a, r_b)
  - w.r.t r_a: is sigmoid(Delta_{ab}) (1 - sigmoid(Delta_{ab})) (-1/tau)
  - w.r.t r_b: is sigmoid(Delta_{ab}) (1 - sigmoid(Delta_{ab})) (1/tau).

which makes the losses easy to compute for any batch of votes.

We call this loss the "sigmoid loss", and it is implemented in the
SoftCondorcetOptimizer class. There is also the Fenchel-Young loss, as described
in Section 3.3 of the paper, which uses a similar gradient descent form but
optimizes a different loss based on perturbed optimizers in machine learning.
The optimizer using the Fenchel-Young loss is implemented in the
FenchelYoungOptimizer class.

Note: this python implementation was the one used for the results in the
original paper. For a faster version, see the C++ implementation in
evaluation/soft_condorcet_optimization.h which is exposed via python bindings
in pyspiel (for an example use from Python, see voting/examples/atari.py).
"""

import abc
import collections
from absl import logging
import numpy as np
from open_spiel.python.voting import base


class Optimizer(abc.ABC):
  """Optimizer without a gradient."""

  def __init__(
      self,
      profile: base.PreferenceProfile,
      batch_size: int,
      rating_lower_bound: float,
      rating_upper_bound: float,
      compute_norm_freq: int,
      initial_param_noise: float,
      verbose: bool = False,
  ):
    self._verbose = verbose
    self._profile = profile
    # Ungroup the profile (make all the votes have weight 1) to make it easier
    # to sample from when assembling batches.
    self._profile.ungroup()
    self._num_alternatives = self._profile.num_alternatives()
    if rating_upper_bound <= rating_lower_bound:
      raise ValueError(
          f"Upper bound ({rating_upper_bound}) must be higher than lower"
          f" bound ({rating_lower_bound})."
      )

    self._rating_ub = rating_upper_bound
    self._rating_lb = rating_lower_bound
    self._batch_size = batch_size
    self._compute_norm_freq = compute_norm_freq
    midpoint_rating = (
        self._rating_ub - self._rating_lb
    ) / 2.0 + self._rating_lb
    self._ratings = np.ones(self._profile.num_alternatives(), dtype=np.float32)
    self._ratings.fill(midpoint_rating)
    self._initial_noise = np.zeros(self._num_alternatives, dtype=np.float32)
    if initial_param_noise > 0.0:
      self._initial_noise = (
          np.random.rand(self._num_alternatives) * initial_param_noise
      )
      self._ratings = self._ratings + self._initial_noise
    self._avg_l2_grad_norm = 0
    self._avg_l1_sum_grad_norm = 0
    self._total_iterations = 0

  @property
  def ratings(self) -> np.ndarray:
    return self._ratings

  @property
  def initial_noise(self) -> np.ndarray:
    return self._initial_noise

  @property
  def total_iterations(self) -> int:
    return self._total_iterations

  @property
  def avg_l2_grad_norm(self) -> float:
    return self._avg_l2_grad_norm

  @property
  def avg_l1_sum_grad_norm(self) -> float:
    return self._avg_l1_sum_grad_norm

  def _gradient(self, ratings: np.ndarray, batch: np.ndarray) -> np.ndarray:
    raise NotImplementedError()

  def step(self, learning_rate: float, batch: np.ndarray) -> np.ndarray:
    """Applies one step of gradient descent on the batch.

    Args:
      learning_rate: a step size for the update.
      batch: the batch of votes (integer indices)

    Returns:
      gradient: the gradient over all parameters.
    """
    gradient = self._gradient(self._ratings, batch)
    self._ratings = self._ratings - learning_rate * gradient
    self._ratings = np.clip(self._ratings, self._rating_lb, self._rating_ub)
    return gradient

  def ranking(self) -> base.PreferenceList:
    """Return a sorted list by decreasing rating."""
    sorted_indices = np.argsort(-self._ratings)
    return [self._profile.alternatives[i] for i in sorted_indices]

  def run_solver(self,
                 iterations: int = 1000,
                 learning_rate: float = 0.01) -> tuple[np.ndarray,
                                                       base.PreferenceList]:
    """Soft Condorcet optimizer."""

    l1_sum_norms = []
    l2_norms = []
    batch = np.arange(self._profile.num_votes(), dtype=int)
    for i in range(iterations):
      self._total_iterations += 1
      if self._batch_size > 0:
        # SGD case: Sample a batch of votes.
        batch = np.random.randint(self._profile.num_votes(),
                                  size=self._batch_size)
      gradient = self.step(learning_rate, batch)
      l2_norms.append(np.linalg.norm(gradient))
      l1_sum_norms.append(np.absolute(gradient).sum())
      if (i - 1) % self._compute_norm_freq == 0:
        self._avg_l1_sum_grad_norm = (
            np.asarray(l1_sum_norms).sum() / self._compute_norm_freq)
        self._avg_l2_grad_norm = (
            np.asarray(l2_norms).sum() / self._compute_norm_freq)
        l2_norms = []
        l1_sum_norms = []
        if self._verbose:
          logging.info("L1 gradient norm = %d", self._avg_l1_sum_grad_norm)
          logging.info("L2 gradient norm = %d", self._avg_l2_grad_norm)
    return self._ratings, self.ranking()

  def approximate_posterior(self,
                            num_posterior_samples: int,
                            num_cov_samples: int) -> np.ndarray:
    """Stochastic Gradient Descent as Approximate Bayesian Inference."""

    gradients = []
    for _ in range(num_cov_samples):
      batch = np.random.randint(self._profile.num_votes(),
                                size=self._batch_size)
      gradient = np.asarray(self._gradient(self._ratings, batch))
      gradients.append(gradient)
    gradients = np.stack(gradients, axis=0)

    gradients_centered = gradients - gradients.mean(axis=0, keepdims=True)
    cov = np.dot(gradients_centered.T, gradients_centered) / num_cov_samples
    # cov_factor = np.linalg.cholesky(cov)  # cov = cov_factor.dot(cov_factor.T)

    coeff = 2 * self._batch_size / self._profile.num_votes()
    precon = coeff * np.linalg.pinv(cov)

    samples = []

    sample = np.array(self._ratings)
    for _ in range(num_posterior_samples):
      batch = np.random.randint(self._profile.num_votes(),
                                size=self._batch_size)
      gradient = self._gradient(sample, batch)
      sample -= precon.dot(gradient)
      sample = np.clip(sample, self._rating_lb, self._rating_ub)
      samples.append(sample)

    samples = np.stack(samples, axis=0)

    return samples


class SoftCondorcetOptimizer(Optimizer):
  """Soft Condorcet optimizer."""

  def __init__(
      self,
      profile: base.PreferenceProfile,
      batch_size: int = 0,  # full GD by default
      rating_lower_bound: float = 0.0,
      rating_upper_bound: float = 1000.0,
      compute_norm_freq: int = 1000,
      initial_param_noise: float = 0.0,
      temperature: float = 1.0,
  ):
    super().__init__(
        profile,
        batch_size,
        rating_lower_bound,
        rating_upper_bound,
        compute_norm_freq,
        initial_param_noise,
    )
    self._temperature = temperature

  def _gradient(self, ratings: np.ndarray, batch: np.ndarray) -> np.ndarray:
    """Compute the gradient of a batch of data. Explained above."""

    alt_idx = self._profile.alternatives_dict
    wins_dict = collections.defaultdict(lambda: 0)
    grad = np.zeros(self._profile.num_alternatives(), dtype=np.float32)
    for idx in batch:
      vote = self._profile.votes[idx]
      vote_len = len(vote.vote)
      for i in range(vote_len):
        for j in range(i+1, vote_len):
          wins_dict[(vote.vote[i], vote.vote[j])] += vote.weight
    # print(dict(wins_dict))
    for alt_tuple, weight in wins_dict.items():
      alt_a, alt_b = alt_tuple
      a_idx = alt_idx[alt_a]
      b_idx = alt_idx[alt_b]
      # print(f"{alt_a} ({a_idx}) {alt_b} ({b_idx}) {weight}")
      delta_ab = ((ratings[b_idx] - ratings[a_idx])
                  / self._temperature)
      sigma_ab = 1.0 / (1.0 + np.exp(-delta_ab))
      grad[a_idx] -= (weight * sigma_ab * (1.0 - sigma_ab)
                      / self._temperature)
      grad[b_idx] += (weight * sigma_ab * (1.0 - sigma_ab)
                      / self._temperature)
    grad /= len(batch)
    return grad


class FenchelYoungOptimizer(Optimizer):
  """Replace gradient by Fenchel Young loss gradient."""

  def __init__(
      self,
      profile: base.PreferenceProfile,
      batch_size: int = 0,  # full GD by default
      rating_lower_bound: float = 0.0,
      rating_upper_bound: float = 1000.0,
      compute_norm_freq: int = 1000,
      initial_param_noise: float = 0.0,
      sigma: float = 100.0,
  ):
    super().__init__(
        profile,
        batch_size,
        rating_lower_bound,
        rating_upper_bound,
        compute_norm_freq,
        initial_param_noise,
    )
    self._sigma = sigma

  def _gradient(self, ratings: np.ndarray, batch: np.ndarray) -> np.ndarray:
    """Compute FY gradient y_eps - y."""
    alt_idx = self._profile.alternatives_dict
    grad = np.zeros(self._profile.num_alternatives(), dtype=np.float32)
    for idx in batch:
      vote = self._profile.votes[idx]
      if vote.weight != 1:
        raise ValueError("Fenchel Young Optimizer only works with weight 1.")
      vote_len = len(vote.vote)
      target_ranking = np.arange(vote_len).astype(np.float32)
      player_ids = [alt_idx[a] for a in vote.vote]
      gumbel_noise = np.random.gumbel(loc=0.0, scale=1.0, size=vote_len).astype(
          np.float32
      )
      # Sample one perturbed ranking. Could do averages of multiple.
      predicted_ratings = ratings[player_ids] + gumbel_noise * self._sigma
      # Randomize tie-breaking by shuffling and unshuffling.
      shuffled = np.random.permutation(len(player_ids))
      unshuffle = np.argsort(shuffled)
      predicted_ranking = np.argsort(np.argsort(-predicted_ratings[shuffled]))[
          unshuffle
      ].astype(np.float32)

      local_grad = predicted_ranking - target_ranking
      # Flipping the sign due to argmin in ranking: Since the loss was derived
      # for f(x)=argmax x, and g(x)=argmin x=f(-x), the gradient g'=-f'(-x).
      grad[player_ids] += -local_grad
    grad /= len(batch)
    return grad



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Regret Matching algorithm for MFGs."""

from typing import Optional

import numpy as np

from open_spiel.python.mfg.algorithms import utils
from open_spiel.python.mfg.algorithms.regret import regret_minimizer


def regret_matching(regrets):
  regrets = np.array(regrets)
  regret_plus = regrets * (regrets > 0.0)
  regrets_sum = np.sum(regret_plus, axis=-1)
  regret_plus[regrets_sum > 0.0, :] = regret_plus[
      regrets_sum > 0.0, :
  ] / regrets_sum[regrets_sum > 0.0].reshape(-1, 1)
  regret_plus[regrets_sum <= 0.0, :] = (
      np.ones_like(regret_plus[regrets_sum <= 0.0, :]) / regret_plus.shape[-1]
  )
  return regret_plus


class RegretMatching(regret_minimizer.RegretMinimizer):
  """Base class for Regret Minimizers.

  Implements base functions for regret minimizers to implement.

  Attributes:
    _game: Pyspiel game.
    _regret_steps_per_step: Number of regret steps per `step` call (Maximum
      number in case `stop_early` is true)
    _rho_tol: If `_compress_nus` is true, minimum probability threshold (
      Probabilities below `rho_tol` will be filtered out).
    _compress_nus: Whether to compress nus (Remove nus with low selection
      probability) or not.
    _compress_lbd: Penalty term in L1 minimization when compressing nus.
    _stop_early: Whether to stop regret computation when average regret is lower
      than `_stop_regret_threshold` or to keep going until
      `_regret_steps_per_step` steps have been accomplished.
    _stop_regret_threshold: If `stop_early` is true, average regret threshold
      under which the algorithm will stop.
    _policies: List of Policies
    _value_estimator: Value estimation function.
    _value_estimation_n: Number of runs to average _value_estimator's result on.
  """

  def __init__(
      self,
      game,
      policies,
      eta: Optional[float] = None,
      regret_steps_per_step: int = 1,
      rho_tol: float = 1e-4,
      compress_nus: bool = True,
      compress_lbd: float = 0.0,
      compress_every: int = 1,
      stop_early: bool = True,
      stop_regret_threshold: float = 1e-3,
      value_estimator=utils.sample_value,
      value_estimation_n: int = 1,
      compute_internal_regret: bool = False,
  ):
    super().__init__(
        game,
        policies,
        regret_steps_per_step=regret_steps_per_step,
        rho_tol=rho_tol,
        compress_nus=compress_nus,
        compress_every=compress_every,
        compress_lbd=compress_lbd,
        stop_early=stop_early,
        stop_regret_threshold=stop_regret_threshold,
        value_estimator=value_estimator,
        value_estimation_n=value_estimation_n,
        compute_internal_regret=compute_internal_regret,
    )

    if self._compute_internal_regret:
      self._regrets = np.zeros((len(policies), len(policies)))
    else:
      self._regrets = np.zeros(len(policies))
    self._p = np.ones(len(policies)) / (1.0 * len(policies))

  def get_all_action_regrets(self):
    assert self._compute_internal_regret
    return [
        regret_matching(np.sum(action_regret, axis=0))
        for action_regret in self._regrets
    ]

  def compute_last_regret(self, nu, reward):
    reward = np.array(reward)
    if self._compute_internal_regret:
      weighted_rewards = nu.reshape(-1, 1) * reward.reshape(1, -1)
      on_policy_values = np.sum(
          regret_matching(self._regrets) * weighted_rewards,
          axis=-1,
          keepdims=True,
      )
      return weighted_rewards - on_policy_values
    else:
      on_policy_value = np.sum(np.array(nu) * np.array(reward))
      return reward - on_policy_value

  def update_regret(self, nu, reward):
    self._regrets += self.compute_last_regret(nu, reward)

  def get_all_w_nus(self):
    assert self._compute_internal_regret
    return regret_matching(self._regrets)

  def get_nu(self):
    if self._compute_internal_regret:
      return np.sum(
          self._p.reshape(-1, 1) * regret_matching(self._regrets), axis=0
      )
    else:
      return regret_matching(self._regrets)

  def step(self, welfare_bonus=0.0):
    rewards = np.zeros(len(self._policies))
    nu = self.get_nu()
    assert np.all(nu >= 0.0) and (np.abs(np.sum(nu) - 1) < 1e-8)
    self._nus.append(nu)
    self._nu_weights.append(1.0)

    mu = utils.MixedDistribution(self._policy_mus, nu)
    for _ in range(self._value_estimation_n):
      for index, policy in enumerate(self._policies):
        rewards[index] += self._value_estimator(policy, mu, self._game)
    rewards /= self._value_estimation_n

    welfare = np.sum(np.array(rewards) * np.array(nu))

    self._rewards.append(rewards + welfare_bonus * welfare * nu)
    self._true_rewards.append(rewards)

    self.update_regret(nu, rewards + welfare_bonus * welfare * nu)
    if self._compute_internal_regret:
      self.compute_p()

  def reset(self, policies):
    """Restart the bandit with new policies."""
    self._p = np.ones(len(policies)) / (1.0 * len(policies))
    self._policies = policies
    self._nus = []
    self._rewards = []
    self._true_rewards = []
    if self._compute_internal_regret:
      self._regrets = np.zeros((len(policies), len(policies)))
    else:
      self._regrets = np.zeros(len(policies))
    self._policy_mus = []
    self._nu_weights = []
    self.update_policy_mus()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example to traverse an entire game tree."""

from absl import app
from absl import flags

from open_spiel.python import games  # pylint: disable=unused-import
import pyspiel

_GAME_STRING = flags.DEFINE_string(
    "game_string", "tic_tac_toe", "Name of the game"
)


class GameStats:
  num_states: int = 0
  num_chance_nodes: int = 0
  num_decision_nodes: int = 0
  num_simultaneous_nodes: int = 0
  num_terminals: int = 0
  info_state_dict: dict[str, list[int]] = {}

  def __str__(self):
    return (f"Number of states {self.num_states} \n" +
            f"Number of chance nodes {self.num_chance_nodes} \n" +
            f"Number of decision nodes {self.num_decision_nodes} \n" +
            f"Number of simultaneous nodes {self.num_simultaneous_nodes} \n" +
            f"Number of terminals {self.num_terminals} \n")


def traverse_game_tree(game: pyspiel.Game,
                       state: pyspiel.State,
                       game_stats: GameStats):
  """Traverses the game tree, collecting information about the game."""

  if state.is_terminal():
    game_stats.num_terminals += 1
  elif state.is_chance_node():
    game_stats.num_chance_nodes += 1
    for outcome in state.legal_actions():
      child = state.child(outcome)
      traverse_game_tree(game, child, game_stats)
  elif state.is_simultaneous_node():
    game_stats.num_simultaneous_nodes += 1
    # Using joint actions for convenience. Can use legal_actions(player) to
    # and state.apply_actions when walking over individual players
    for joint_action in state.legal_actions():
      child = state.child(joint_action)
      traverse_game_tree(game, child, game_stats)
  else:
    game_stats.num_decision_nodes += 1
    legal_actions = state.legal_actions()
    if game.get_type().provides_information_state_string:
      game_stats.info_state_dict[
          state.information_state_string()] = legal_actions
    for action in state.legal_actions():
      # print(f"Decision node: \n {state}")
      # print(f"Taking action {action} ({state.action_to_string(action)}")
      child = state.child(action)
      traverse_game_tree(game, child, game_stats)


def main(_):
  game = pyspiel.load_game(_GAME_STRING.value)
  game_stats = GameStats()
  state = game.new_initial_state()
  traverse_game_tree(game, state, game_stats)
  print(game_stats)
  # for info_state_string in game_stats.info_state_dict:
  #   print(info_state_string)
  #   # print(game_stats.info_state_dict[info_state_string])  # legal actions


if __name__ == "__main__":
  app.run(main)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
import numpy as np
from open_spiel.python.coalitional_games import basic_games
from open_spiel.python.coalitional_games import least_core_lp


SEED = 817346817


class LeastCoreLPTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    np.random.seed(SEED)

  def test_ice_cream_example_full_lp(self):
    """Solve the full LP."""
    game = basic_games.IceCreamGame()
    imputation, epsilon = least_core_lp.solve_least_core_lp(
        game, least_core_lp.add_all_constraints)
    self.assertAlmostEqual(imputation.sum(), 1000.0, delta=1e-5)
    self.assertGreater(imputation.all(), 0.0)
    self.assertLess(epsilon, 1e-6)

  def test_ice_cream_example_uniform_sample_lp(self):
    """Solve the LP with 20 uniformly sampled constraints."""
    game = basic_games.IceCreamGame()
    cons_func = least_core_lp.make_uniform_sampling_constraints_function(20)
    imputation, epsilon = least_core_lp.solve_least_core_lp(game, cons_func)
    self.assertAlmostEqual(imputation.sum(), 1000.0, delta=1e-5)
    self.assertGreater(imputation.all(), 0.0)
    self.assertLess(epsilon, 1e-6)


if __name__ == "__main__":
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for least-core lagrangian calculations."""

from absl.testing import absltest
from ml_collections import config_dict as configdict
import numpy as np

from open_spiel.python.coalitional_games import basic_games
from open_spiel.python.coalitional_games import least_core_lagrangian


SEED = 817346817


def get_alg_config():
  """Get configuration for botched trades experiment."""
  alg_config = configdict.ConfigDict()

  alg_config.init = configdict.ConfigDict()
  alg_config.init.lr_primal = 1e-2
  alg_config.init.lr_dual = 1e-2

  alg_config.solve = configdict.ConfigDict()
  alg_config.solve.batch_size = 2**3
  alg_config.solve.mu_init = 1000
  alg_config.solve.gamma = 1e-8
  alg_config.solve.n_iter = 110_000
  alg_config.solve.seed = 0
  alg_config.solve.save_every = 10_000

  alg_config.eval = configdict.ConfigDict()
  alg_config.eval.evaluation_iterations = 2**3

  return alg_config


class LeastCoreLagrangianTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    np.random.seed(SEED)
    self.config = get_alg_config()

  def test_ice_cream_example_full_lagrangian(self):
    """Solve the least core Lagrangian."""
    game = basic_games.IceCreamGame()
    least_core_value = least_core_lagrangian.compute_least_core_value(
        game, self.config)
    imputation = least_core_value.payoff
    epsilon = least_core_value.lcv
    self.assertAlmostEqual(imputation.sum(), 1000.0, places=3)
    self.assertGreater(imputation.all(), -1e-10)
    self.assertLess(epsilon, 1e-6)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of Fictitious Play from Perrin & al.

Reference: https://arxiv.org/abs/2007.03458.
As presented, the Fictitious Play algorithm provides a robust approximation
scheme for Nash equilibrium by iteratively computing the best response
against the distribution induced by the average of the past best responses.
The provided formulation of Deep Fictitious Play mirrors this procedure,
but substitutes out the exact best reponse computation with an approximation
of best response values through a Reinforcement Learning approach (where
the RL method in question is a user-determined parameter for each iteration).

Policy is initialized to uniform policy.
Each iteration:
 1. Compute best response against policy
 2. Update policy as weighted average of best response and current policy
    (default learning rate is 1 / num_iterations + 1).

To use fictitious play one should initialize it and run multiple iterations:
fp = FictitiousPlay(game)
for _ in range(num_iterations):
  fp.iteration()
policy = fp.get_policy()
"""

import math
from typing import List, Optional

from open_spiel.python import policy as policy_std
from open_spiel.python.mfg import distribution as distribution_std
from open_spiel.python.mfg import value
from open_spiel.python.mfg.algorithms import best_response_value
from open_spiel.python.mfg.algorithms import distribution
from open_spiel.python.mfg.algorithms import greedy_policy
from open_spiel.python.mfg.algorithms import policy_value
from open_spiel.python.mfg.algorithms import softmax_policy
import pyspiel


class MergedPolicy(policy_std.Policy):
  """Merge several policies."""

  def __init__(
      self,
      game,
      player_ids: List[int],
      policies: List[policy_std.Policy],
      distributions: List[distribution_std.Distribution],
      weights: List[float],
  ):
    """Initializes the merged policy.

    Args:
      game: The game to analyze.
      player_ids: list of player ids for which this policy applies; each should
        be in the range 0..game.num_players()-1.
      policies: A `List[policy_std.Policy]` object.
      distributions: A `List[distribution_std.Distribution]` object.
      weights: A `List[float]` object. The elements should sum to 1.
    """
    super().__init__(game, player_ids)
    self._policies = policies
    self._distributions = distributions
    self._weights = weights
    assert len(policies) == len(distributions), (
        f'Length mismatch {len(policies)} != {len(distributions)}')
    assert len(policies) == len(weights), (
        f'Length mismatch {len(policies)} != {len(weights)}')
    assert math.isclose(
        sum(weights),
        1.0), (f'Weights should sum to 1, but instead sum to {sum(weights)}')

  def action_probabilities(self, state, player_id=None):
    action_prob = []
    legal = state.legal_actions()
    num_legal = len(legal)
    for a in legal:
      merged_pi = 0.0
      norm_merged_pi = 0.0
      for p, d, w in zip(self._policies, self._distributions, self._weights):
        try:
          merged_pi += w * d(state) * p(state)[a]
          norm_merged_pi += w * d(state)
        except (KeyError, ValueError):
          # This happens when the state was not observed in the merged
          # distributions or policies.
          pass
      if norm_merged_pi > 0.0:
        action_prob.append((a, merged_pi / norm_merged_pi))
      else:
        action_prob.append((a, 1.0 / num_legal))
    return dict(action_prob)


class FictitiousPlay(object):
  """Computes the value of a specified strategy."""

  def __init__(self,
               game: pyspiel.Game,
               lr: Optional[float] = None,
               temperature: Optional[float] = None):
    """Initializes the greedy policy.

    Args:
      game: The game to analyze.
      lr: The learning rate of mirror descent. If None, at iteration i it will
        be set to 1/i.
      temperature: If set, then instead of the greedy policy a softmax policy
        with the specified temperature will be used to update the policy at each
        iteration.
    """
    self._game = game
    self._lr = lr
    self._temperature = temperature
    self._policy = policy_std.UniformRandomPolicy(self._game)

    self._correlating_policy = self._policy
    self._distribution = distribution.DistributionPolicy(
        self._game, self._correlating_policy
    )
    self._fp_step = 0

  def get_policy(self):
    return self._policy

  def get_correlating_policy(self):
    return self._policy

  def get_correlating_distribution(self):
    return distribution.DistributionPolicy(self._game, self._policy)

  def iteration(self, br_policy=None, learning_rate=None):
    """Returns a new `TabularPolicy` equivalent to this policy.

    Args:
      br_policy: Policy to compute the best response value for each iteration.
        If none provided, the exact value is computed.
      learning_rate: The learning rate.
    """
    self._fp_step += 1

    distrib = distribution.DistributionPolicy(self._game, self._policy)

    if br_policy:
      br_value = policy_value.PolicyValue(self._game, distrib, br_policy)
    else:
      br_value = best_response_value.BestResponse(
          self._game, distrib, value.TabularValueFunction(self._game))

    # Policy is either greedy or softmax with respect to the best response if
    # temperature is specified.
    player_ids = list(range(self._game.num_players()))
    if self._temperature is None:
      pi = greedy_policy.GreedyPolicy(self._game, player_ids, br_value)
    else:
      pi = softmax_policy.SoftmaxPolicy(self._game, player_ids,
                                        self._temperature, br_value)
    pi = pi.to_tabular()

    distrib_pi = distribution.DistributionPolicy(self._game, pi)

    if learning_rate:
      weight = learning_rate
    else:
      weight = self._lr if self._lr else 1.0 / (self._fp_step + 1)

    self._correlating_policy = pi
    self._distribution = distrib_pi

    if math.isclose(weight, 1.0):
      self._policy = pi
    else:
      self._policy = MergedPolicy(self._game, player_ids, [self._policy, pi],
                                  [distrib, distrib_pi],
                                  [1.0 - weight, weight]).to_tabular()
